import torch

from keras.src import tree
from keras.src.backend.torch.core import convert_to_tensor
from keras.src.backend.torch.core import get_device


def rnn(
    step_function,
    inputs,
    initial_states,
    go_backwards=False,
    mask=None,
    constants=None,
    unroll=False,
    input_length=None,
    time_major=False,
    zero_output_for_mask=False,
    return_all_outputs=True,
):
    input_length = input_length or inputs.shape[1]

    def swap_batch_timestep(input_t):
        # Swap the batch and timestep dim for the incoming tensor.
        axes = list(range(len(input_t.shape)))
        axes[0], axes[1] = 1, 0
        return torch.permute(input_t, axes)

    if not time_major:
        inputs = tree.map_structure(swap_batch_timestep, inputs)

    flattened_inputs = tree.flatten(inputs)
    time_steps = flattened_inputs[0].shape[0]
    time_steps_t = time_steps

    if mask is not None:
        if mask.dtype != torch.bool:
            mask = mask.type(torch.bool)
        if len(mask.shape) == 2:
            mask = torch.unsqueeze(mask, -1)
        if not time_major:
            mask = swap_batch_timestep(mask)

    if constants is None:
        constants = []

    def _expand_mask(mask_t, input_t, fixed_dim=1):
        if tree.is_nested(mask_t):
            raise ValueError(
                f"mask_t is expected to be tensor,\
                  but got {mask_t}"
            )
        if tree.is_nested(input_t):
            raise ValueError(
                f"input_t is expected to be tensor,\
                  but got {input_t}"
            )
        rank_diff = len(input_t.shape) - len(mask_t.shape)
        for _ in range(rank_diff):
            mask_t = torch.unsqueeze(mask_t, -1)
        multiples = [1] * fixed_dim + list(input_t.shape[fixed_dim:])
        return torch.tile(mask_t, multiples)

    if unroll:
        if not time_steps:
            raise ValueError("Unrolling requires a fixed number of timesteps.")
        states = tuple(initial_states)
        successive_states = []
        successive_outputs = []

        # Process the input tensors. The input tensor need to be split on the
        # time_step dim, and reverse if go_backwards is True. In the case of
        # nested input, the input is flattened and then transformed
        # individually.  The result of this will be a tuple of lists, each of
        # the item in tuple is list of the tensor with shape (batch, feature)
        def _process_single_input_t(input_t):
            input_t = torch.unbind(input_t)  # unstack for time_step dim
            if go_backwards:
                input_t = input_t[::-1]
            return input_t

        if tree.is_nested(inputs):
            processed_input = tree.map_structure(
                _process_single_input_t, inputs
            )  # noqa: E501
        else:
            processed_input = (_process_single_input_t(inputs),)

        def _get_input_tensor(time):
            inp = [t_[time] for t_ in processed_input]
            return tree.pack_sequence_as(inputs, inp)

        if mask is not None:
            mask_list = torch.unbind(mask)
            if go_backwards:
                mask_list = torch.flip(mask_list, dims=mask_list.shape)

            for i in range(time_steps):
                inp = _get_input_tensor(i)
                mask_t = mask_list[i]
                output, new_states = step_function(
                    inp, tuple(states) + tuple(constants)
                )
                tiled_mask_t = _expand_mask(mask_t, output)

                if not successive_outputs:
                    prev_output = torch.zeros_like(output)
                else:
                    prev_output = successive_outputs[-1]

                output = torch.where(tiled_mask_t, output, prev_output)

                flat_states = tree.flatten(states)
                flat_new_states = tree.flatten(new_states)
                tiled_mask_t = tuple(
                    _expand_mask(mask_t, s) for s in flat_states
                )  # noqa: E501
                flat_final_states = tuple(
                    torch.where(m, s, ps)
                    for m, s, ps in zip(
                        tiled_mask_t, flat_new_states, flat_states
                    )  # noqa: E501
                )
                states = tree.pack_sequence_as(states, flat_final_states)

                if return_all_outputs:
                    successive_outputs.append(output)
                    successive_states.append(states)
                else:
                    successive_outputs = [output]
                    successive_states = [states]
            last_output = successive_outputs[-1]
            new_states = successive_states[-1]
            outputs = torch.stack(successive_outputs)

            if zero_output_for_mask:
                last_output = torch.where(
                    _expand_mask(mask_list[-1], last_output),
                    last_output,
                    torch.zeros_like(last_output),
                )
                outputs = torch.where(
                    _expand_mask(mask, outputs, fixed_dim=2),
                    outputs,
                    torch.zeros_like(outputs),
                )

        else:  # mask is None
            for i in range(time_steps):
                inp = _get_input_tensor(i)
                output, states = step_function(
                    inp, tuple(states) + tuple(constants)
                )  # noqa: E501
                if return_all_outputs:
                    successive_outputs.append(output)
                    successive_states.append(states)
                else:
                    successive_outputs = [output]
                    successive_states = [states]
            last_output = successive_outputs[-1]
            new_states = successive_states[-1]
            outputs = torch.stack(successive_outputs)

    else:  # Unroll == False
        states = tuple(initial_states)

        # Create input tensor array, if the inputs is nested tensors, then it
        # will be flattened first, and tensor array will be created one per
        # flattened tensor.

        input_ta = tuple(
            (
                list(torch.unbind(input_))
                if not go_backwards
                else list(torch.unbind(torch.flip(input_, [0])))
            )
            for input_ in flattened_inputs
        )

        # Get the time(0) input and compute the output for that.
        input_time_zero = tree.pack_sequence_as(
            inputs, [inp[0] for inp in flattened_inputs]
        )
        # output_time_zero is used to determine the cell output shape.
        output_time_zero, _ = step_function(
            input_time_zero, tuple(initial_states) + tuple(constants)
        )

        output_ta_size = time_steps_t if return_all_outputs else 1
        output_ta = []
        for out in tree.flatten(output_time_zero):
            out_list = list(out)
            if len(out) < output_ta_size:
                out_list.extend([[]] * (output_ta_size - len(out)))
            output_ta.append(out_list)

        time = torch.tensor(0, dtype=torch.int32)

        if input_length is None:
            max_iterations = time_steps_t
        else:
            if hasattr(input_length, "__len__"):
                input_length = convert_to_tensor(input_length)
                max_iterations = torch.max(input_length)
            else:
                max_iterations = input_length

        if mask is not None:
            if go_backwards:
                mask = torch.flip(mask, [0])

            mask_ta = list(torch.unbind(mask))

            def masking_fn(time):
                return mask_ta[time]

            def compute_masked_output(mask_t, flat_out, flat_mask):
                tiled_mask_t = tuple(
                    _expand_mask(mask_t, o, fixed_dim=len(mask_t.shape))
                    for o in flat_out
                )
                return tuple(
                    torch.where(m, o, fm)
                    for m, o, fm in zip(tiled_mask_t, flat_out, flat_mask)
                )

        elif isinstance(input_length, torch.Tensor):
            if go_backwards:
                max_len = torch.max(input_length, dim=0)
                if isinstance(max_len, torch.return_types.max):
                    max_len = max_len[0]
                rev_input_length = torch.subtract(max_len - 1, input_length)

                def masking_fn(time):
                    return torch.less(rev_input_length, time)

            else:

                def masking_fn(time):
                    return torch.greater(input_length, time)

            def compute_masked_output(mask_t, flat_out, flat_mask):
                return tuple(
                    torch.where(mask_t, o, zo)
                    for (o, zo) in zip(flat_out, flat_mask)  # noqa: E501
                )

        else:
            masking_fn = None

        if masking_fn is not None:
            # Mask for the T output will be base on the output of T - 1. In the
            # case T = 0, a zero filled tensor will be used.
            flat_zero_output = tuple(
                torch.zeros_like(o) for o in tree.flatten(output_time_zero)
            )

            def _step(time, output_ta_t, prev_output, *states):
                """RNN step function.

                Args:
                    time: Current timestep value.
                    output_ta_t: TensorArray.
                    prev_output: tuple of outputs from time - 1.
                    *states: List of states.

                Returns:
                    Tuple: `(time + 1, output_ta_t, output) + tuple(new_states)`
                """
                current_input = tuple(ta[time] for ta in input_ta)
                # maybe set shape.
                current_input = tree.pack_sequence_as(inputs, current_input)
                mask_t = masking_fn(time)
                output, new_states = step_function(
                    current_input, tuple(states) + tuple(constants)
                )
                # mask output
                flat_output = tree.flatten(output)
                flat_mask_output = (
                    flat_zero_output
                    if zero_output_for_mask
                    else tree.flatten(prev_output)
                )
                flat_new_output = compute_masked_output(
                    mask_t, flat_output, flat_mask_output
                )

                # mask states
                flat_state = tree.flatten(states)
                flat_new_state = tree.flatten(new_states)
                flat_final_state = compute_masked_output(
                    mask_t, flat_new_state, flat_state
                )
                new_states = tree.pack_sequence_as(new_states, flat_final_state)  # noqa: E501

                ta_index_to_write = time if return_all_outputs else 0
                for ta, out in zip(output_ta_t, flat_new_output):
                    ta[ta_index_to_write] = out

                return (time + 1, output_ta_t, tuple(flat_new_output)) + tuple(
                    new_states
                )

            it = 0
            output_ta_t, new_states, prev_output = (
                output_ta,
                states,
                flat_zero_output,
            )
            while time < time_steps_t and it < max_iterations:
                final_outputs = _step(
                    time, output_ta_t, prev_output, *new_states
                )  # noqa: E501
                time, output_ta_t, prev_output = final_outputs[:3]
                new_states = final_outputs[3:]
                it += 1

        else:

            def _step(time, output_ta_t, *states):
                """RNN step function.

                Args:
                    time: Current timestep value.
                    output_ta_t: TensorArray.
                    *states: List of states.

                Returns:
                    Tuple: `(time + 1,output_ta_t) + tuple(new_states)`
                """
                current_input = tuple(ta[time] for ta in input_ta)
                current_input = tree.pack_sequence_as(inputs, current_input)
                output, new_states = step_function(
                    current_input, tuple(states) + tuple(constants)
                )
                flat_new_state = tree.flatten(new_states)

                flat_output = tree.flatten(output)
                ta_index_to_write = time if return_all_outputs else 0
                for ta, out in zip(output_ta_t, flat_output):
                    ta[ta_index_to_write] = out

                new_states = tree.pack_sequence_as(
                    initial_states, flat_new_state
                )  # noqa: E501
                return (time + 1, output_ta_t) + tuple(new_states)

            it = 0
            output_ta_t = output_ta
            new_states = states
            while time < time_steps_t and it < max_iterations:
                final_outputs = _step(time, output_ta_t, *new_states)
                time, output_ta_t = final_outputs[:2]
                new_states = final_outputs[2:]
                it += 1

        def _stack(tensor_list):
            max_ndims = max([t.ndim for t in tensor_list])
            max_list = []
            for i, t in enumerate(tensor_list):
                if t.ndim == max_ndims:
                    max_list.append(t)
            return torch.stack(max_list)

        output_ta = final_outputs[1]

        outputs = tuple(_stack(o) for o in output_ta)
        last_output = tuple(o[-1] for o in outputs)

        outputs = tree.pack_sequence_as(output_time_zero, outputs)
        last_output = tree.pack_sequence_as(output_time_zero, last_output)

    if not time_major:
        outputs = tree.map_structure(swap_batch_timestep, outputs)

    return last_output, outputs, new_states


def _is_sequence_right_padded(mask):
    """Check the mask tensor and see if it right padded.

    cuDNN uses the sequence length param to skip the tailing
    timestep. If the data is left padded, or not a strict right padding (has
    masked value in the middle of the sequence), then cuDNN won't work
    properly in those cases.

    Left padded data: [[False, False, True, True, True]].
    Right padded data: [[True, True, True, False, False]].
    Mixture of mask/unmasked data: [[True, False, True, False, False]].

    Note that for the mixed data example above, the actually data RNN should see
    are those 2 Trues (index 0 and 2), the index 1 False should be ignored and
    not pollute the internal states.

    Args:
        mask: the Boolean tensor with shape [batch, timestep]

    Returns:
        boolean scalar tensor, whether the mask is strictly right padded.
    """
    # Get max sequence length
    max_seq_length = mask.shape[1]
    # Count True values in each sequence
    count_of_true = torch.sum(mask, dim=1)
    # Create right padded mask
    batch_size = mask.shape[0]
    indices = torch.arange(max_seq_length, device=mask.device).repeat(
        batch_size, 1
    )  # noqa: E501
    right_padded_mask = indices < count_of_true.unsqueeze(1)
    return torch.all(mask == right_padded_mask)


def _has_fully_masked_sequence(mask):
    """Check if input sequence contains any fully masked data.

    cuDNN kernel will error out if the input sequence contains any fully masked
    data. We work around this issue by rerouting the computation to the
    standard kernel until the issue on the cuDNN side has been fixed. For a
    fully masked sequence, it will contain all `False` values. To make it easy
    to check, we invert the boolean and check if any of the sequences has all
    `True` values.

    Args:
        mask: The mask tensor.

    Returns:
        A boolean tensor, `True` if the mask contains a fully masked sequence.
    """
    return torch.any(torch.all(~mask, dim=1))


def _assert_valid_mask(mask):
    # Check if mask is valid for cuDNN
    no_fully_masked = ~_has_fully_masked_sequence(mask)
    is_right_padded = _is_sequence_right_padded(mask)
    valid = no_fully_masked & is_right_padded

    if not valid.item():
        error_message = (
            "You are passing a RNN mask that does not correspond to "
            "right-padded sequences, while using cuDNN, which is not "
            "supported. With cuDNN, RNN masks can only be used for "
            "right-padding, e.g. `[[True, True, False, False]]` would "
            "be a valid mask, but any mask that isn't just contiguous "
            "`True`'s on the left and contiguous `False`'s on the right "
            "would be invalid. You can pass `use_cudnn=False` to your "
            "RNN layer to stop using cuDNN (this may be slower)."
        )
        raise ValueError(error_message)


def _compute_sequence_length_from_mask(mask, batch_first):
    """Calculate the sequence length tensor (1-D) based on the masking tensor.

    The masking tensor is a 2D boolean tensor with shape [batch, timestep]. For
    any timestep that should be masked, the corresponding field will be False.
    Consider the following example:
        a = [[True, True, False, False]
             [True, True, True, False]]
    It is a (2, 4) tensor, and the corresponding sequence length result should
    be 1D tensor with value [2, 3]. Note that the masking tensor must be right
    padded that could be checked by, e.g., `is_sequence_right_padded()`.

    Args:
        mask: Boolean tensor with shape [batch, timestep] or [timestep, batch]
            if time_major=True.
        time_major: Boolean, which indicates whether the mask is time major or
            batch major.

    Returns:
        sequence_length: 1D int32 tensor.
    """
    timestep_index = 0 if not batch_first else 1
    return torch.sum(mask.int(), dim=timestep_index)


def prepare_lstm_params(kernel, recurrent_kernel, bias, device):
    """Prepares Keras LSTM weights for PyTorch's functional LSTM.

    Transposes weight matrices from Keras (input_dim, 4*units) to PyTorch
    (4*units, input_dim) format and returns weight tensors that maintain
    gradient connections.

    Args:
        kernel: The kernel weights tensor with shape (input_dim, 4*units).
        recurrent_kernel: The recurrent kernel weights tensor
            with shape (units, 4*units).
        bias: The bias tensor with shape (4*units,).
        device: The device to place the tensors on.

    Returns:
        A list of weight tensors [weight_ih, weight_hh, bias_ih, bias_hh]
        suitable for torch._VF.lstm.
    """
    # Gate order is the same in Keras and PyTorch: [i, f, c, o]
    # Just need to transpose from (input_dim, 4*units) to (4*units, input_dim)
    weight_ih = kernel.T.contiguous().to(device)
    weight_hh = recurrent_kernel.T.contiguous().to(device)

    hidden_size = recurrent_kernel.shape[0]
    if bias is not None:
        bias_ih = convert_to_tensor(bias).contiguous().to(device)
        bias_hh = torch.zeros(
            4 * hidden_size, dtype=bias_ih.dtype, device=device
        )
    else:
        bias_ih = torch.zeros(
            4 * hidden_size, dtype=kernel.dtype, device=device
        )
        bias_hh = torch.zeros(
            4 * hidden_size, dtype=kernel.dtype, device=device
        )

    return [weight_ih, weight_hh, bias_ih, bias_hh]


def _is_cuda_cudnn_available():
    # We check if the cuda device and drivers are available
    return torch.cuda.is_available() and torch.backends.cudnn.is_available()


def cudnn_ok(
    activation,
    recurrent_activation,
    unroll,
    use_bias=True,
):
    from keras.src import activations
    from keras.src import ops

    return (
        activation in (activations.tanh, torch.tanh, ops.tanh)
        and recurrent_activation
        in (activations.sigmoid, torch.sigmoid, ops.sigmoid)  # noqa: E501
        and not unroll
        and use_bias
        and _is_cuda_cudnn_available()
    )


def lstm(
    inputs,
    initial_state_h,
    initial_state_c,
    mask,
    kernel,
    recurrent_kernel,
    bias,
    activation,
    recurrent_activation,
    return_sequences=False,
    go_backwards=False,
    unroll=False,
    batch_first=True,
):
    # Masking is not supported by either the cuDNN or fallback path;
    # fall back to the generic RNN loop before doing any work.
    if mask is not None:
        raise NotImplementedError

    # Get device from inputs
    device = get_device()

    # Convert to torch tensors (convert_to_tensor unwraps Variables)
    kernel = convert_to_tensor(kernel)
    recurrent_kernel = convert_to_tensor(recurrent_kernel)
    if bias is not None:
        bias = convert_to_tensor(bias)

    # Cast inputs/states to the kernel's dtype so integer inputs are promoted
    # to float and mixed-precision dtypes (e.g. float16) are respected.
    compute_dtype = kernel.dtype
    inputs = convert_to_tensor(inputs).to(compute_dtype)
    initial_state_h = convert_to_tensor(initial_state_h).to(compute_dtype)
    initial_state_c = convert_to_tensor(initial_state_c).to(compute_dtype)

    # Preprocess for go_backwards by flipping the sequence
    if go_backwards:
        seq_dim = 1 if batch_first else 0
        inputs = torch.flip(inputs, dims=[seq_dim])

    # Move all tensors to the same device
    inputs = inputs.to(device)
    initial_state_h = initial_state_h.to(device)
    initial_state_c = initial_state_c.to(device)

    cudnn_supported = cudnn_ok(
        activation,
        recurrent_activation,
        unroll,
        use_bias=bias is not None,
    )

    if cudnn_supported:
        cudnn_inputs = inputs
        if not batch_first:
            cudnn_inputs = inputs.permute(1, 0, 2)
        try:
            last_output, outputs, states = _cudnn_lstm(
                cudnn_inputs,
                initial_state_h,
                initial_state_c,
                kernel,
                recurrent_kernel,
                bias,
                return_sequences=return_sequences,
                device=device,
            )
            if not batch_first:
                outputs = outputs.permute(1, 0, 2)
            return last_output, outputs, states
        except Exception:
            pass

    return _fallback_lstm(
        inputs,
        initial_state_h,
        initial_state_c,
        kernel,
        recurrent_kernel,
        bias,
        activation,
        recurrent_activation,
        return_sequences,
        batch_first,
    )


def _cudnn_lstm(
    inputs,
    initial_state_h,
    initial_state_c,
    kernel,
    recurrent_kernel,
    bias,
    return_sequences,
    device,
):
    # If shape is [batch, hidden]; Make [1, batch, hidden]
    if initial_state_h.dim() == 2:
        initial_state_h = initial_state_h.unsqueeze(0)
        initial_state_c = initial_state_c.unsqueeze(0)
    elif initial_state_h.dim() == 3 and initial_state_h.shape[1] == 1:
        initial_state_h = initial_state_h.permute(1, 0, 2)
        initial_state_c = initial_state_c.permute(1, 0, 2)

    params = prepare_lstm_params(kernel, recurrent_kernel, bias, device)

    # Use functional LSTM to maintain gradient flow through weight tensors
    outputs, (h_n, c_n) = torch._VF.lstm(
        inputs,
        (initial_state_h, initial_state_c),
        params,
        bias is not None,  # has_biases
        1,  # num_layers
        0.0,  # dropout
        torch.is_grad_enabled(),  # training
        False,  # bidirectional
        True,  # batch_first
    )

    h_n = h_n.squeeze(0)
    c_n = c_n.squeeze(0)
    last_output = outputs[:, -1]

    if not return_sequences:
        outputs = last_output.unsqueeze(1)

    return last_output, outputs, [h_n, c_n]


def _fallback_lstm(
    inputs,
    initial_state_h,
    initial_state_c,
    kernel,
    recurrent_kernel,
    bias,
    activation,
    recurrent_activation,
    return_sequences,
    batch_first,
):
    """Pure-torch LSTM with pre-computed input projections.

    Used when cuDNN is not available (CPU, non-standard activations, etc.).
    Pre-computes all input projections in a single matmul across timesteps,
    then only computes the recurrent projection per step.
    """
    if batch_first:
        inputs = inputs.permute(1, 0, 2)  # -> (time, batch, features)

    # Pre-compute input projections: (time, batch, 4*units)
    x_proj = torch.matmul(inputs, kernel)
    if bias is not None:
        x_proj = x_proj + bias

    time_steps = inputs.shape[0]
    h = initial_state_h
    c = initial_state_c
    outputs = []

    for t in range(time_steps):
        z = x_proj[t] + torch.matmul(h, recurrent_kernel)
        z_i, z_f, z_c, z_o = torch.chunk(z, 4, dim=1)

        new_c = recurrent_activation(z_f) * c + recurrent_activation(
            z_i
        ) * activation(z_c)
        new_h = recurrent_activation(z_o) * activation(new_c)

        h = new_h
        c = new_c
        outputs.append(h)

    outputs = torch.stack(outputs, dim=0)  # (time, batch, units)

    if batch_first:
        outputs = outputs.permute(1, 0, 2)  # (batch, time, units)

    last_output = h

    if not return_sequences:
        outputs = last_output.unsqueeze(1 if batch_first else 0)

    return last_output, outputs, [h, c]


def gru(
    inputs,
    initial_state,
    mask,
    kernel,
    recurrent_kernel,
    bias,
    activation,
    recurrent_activation,
    return_sequences=False,
    go_backwards=False,
    unroll=False,
    reset_after=True,
):
    cudnn_supported = cudnn_ok(
        activation,
        recurrent_activation,
        unroll,
        use_bias=bias is not None,
    )

    if not cudnn_supported or not reset_after or mask is not None:
        raise NotImplementedError

    # Get device from inputs
    device = get_device()

    # Convert to torch tensors (convert_to_tensor unwraps Variables)
    kernel = convert_to_tensor(kernel)
    recurrent_kernel = convert_to_tensor(recurrent_kernel)
    if bias is not None:
        bias = convert_to_tensor(bias)

    inputs = convert_to_tensor(inputs, dtype="float32")
    initial_state = convert_to_tensor(initial_state, dtype="float32")

    # Preprocess for go_backwards by flipping the sequence
    if go_backwards:
        inputs = torch.flip(inputs, dims=[1])

    # Move all tensors to the same device
    inputs = inputs.to(device)
    initial_state = initial_state.to(device)

    try:
        return _cudnn_gru(
            inputs,
            initial_state,
            kernel,
            recurrent_kernel,
            bias,
            return_sequences=return_sequences,
            device=device,
        )
    except Exception:
        raise NotImplementedError


def prepare_gru_params(kernel, recurrent_kernel, bias, device):
    """Prepares Keras GRU weights for PyTorch's functional GRU.

    Reorders gates from Keras [z, r, h] to PyTorch [r, z, h] format
    and returns weight tensors that maintain gradient connections.

    Args:
        kernel: The kernel weights tensor with shape (input_dim, 3*units).
        recurrent_kernel: The recurrent kernel weights tensor
            with shape (units, 3*units).
        bias: The bias tensor with shape (2, 3*units) for reset_after=True.
        device: The device to place the tensors on.

    Returns:
        A list of weight tensors [weight_ih, weight_hh, bias_ih, bias_hh]
        suitable for torch._VF.gru.
    """
    # Split Keras weights by gate: [z, r, h]
    z_k, r_k, h_k = torch.chunk(kernel, 3, dim=1)
    z_r, r_r, h_r = torch.chunk(recurrent_kernel, 3, dim=1)

    # Reorder to PyTorch format [r, z, h] and transpose
    weight_ih = torch.cat([r_k, z_k, h_k], dim=1).T.contiguous().to(device)
    weight_hh = torch.cat([r_r, z_r, h_r], dim=1).T.contiguous().to(device)

    if bias is not None:
        # bias shape is (2, 3*units) for reset_after=True
        # Row 0 is input bias, Row 1 is recurrent bias
        z_bi, r_bi, h_bi = torch.chunk(bias[0], 3)
        z_bh, r_bh, h_bh = torch.chunk(bias[1], 3)

        # Reorder to [r, z, h]
        bias_ih = torch.cat([r_bi, z_bi, h_bi]).contiguous().to(device)
        bias_hh = torch.cat([r_bh, z_bh, h_bh]).contiguous().to(device)
    else:
        hidden_size = recurrent_kernel.shape[0]
        bias_ih = torch.zeros(
            3 * hidden_size, dtype=kernel.dtype, device=device
        )
        bias_hh = torch.zeros(
            3 * hidden_size, dtype=kernel.dtype, device=device
        )

    return [weight_ih, weight_hh, bias_ih, bias_hh]


def _cudnn_gru(
    inputs,
    initial_state,
    kernel,
    recurrent_kernel,
    bias,
    return_sequences,
    device,
):
    # If shape is [batch, hidden]; Make [1, batch, hidden]
    if initial_state.dim() == 2:
        initial_state = initial_state.unsqueeze(0)
    # If shape is [batch, 1, hidden]
    elif initial_state.dim() == 3 and initial_state.shape[1] == 1:
        initial_state = initial_state.permute(1, 0, 2)

    params = prepare_gru_params(kernel, recurrent_kernel, bias, device)

    # Use functional GRU to maintain gradient flow through weight tensors
    outputs, h_n = torch._VF.gru(
        inputs,
        initial_state,
        params,
        bias is not None,  # has_biases
        1,  # num_layers
        0.0,  # dropout
        torch.is_grad_enabled(),  # training: must be True for backward pass
        False,  # bidirectional
        True,  # batch_first
    )

    # Reshape hidden state for return
    h_n = h_n.squeeze(0)
    last_output = outputs[:, -1]

    if not return_sequences:
        outputs = last_output.unsqueeze(1)

    return last_output, outputs, [h_n]



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.ops.linalg import cholesky as cholesky
from keras.src.ops.linalg import cholesky_inverse as cholesky_inverse
from keras.src.ops.linalg import det as det
from keras.src.ops.linalg import eig as eig
from keras.src.ops.linalg import eigh as eigh
from keras.src.ops.linalg import inv as inv
from keras.src.ops.linalg import jvp as jvp
from keras.src.ops.linalg import lstsq as lstsq
from keras.src.ops.linalg import lu_factor as lu_factor
from keras.src.ops.linalg import norm as norm
from keras.src.ops.linalg import qr as qr
from keras.src.ops.linalg import solve as solve
from keras.src.ops.linalg import solve_triangular as solve_triangular
from keras.src.ops.linalg import svd as svd



import functools

import jax.experimental.sparse as jax_sparse
import jax.numpy as jnp

from keras.src.utils import jax_utils


def axis_shape_dims_for_broadcast_in_dim(axis, input_shape, insert_dims):
    """Turn the `axis` argument to the arguments needed by `broadcast_in_dim`.

    Args:
        axis: single int or a tuple of ints for the axis argument. The list of
          dimensions to reduce or insert.
        input_shape: the shape of the input as a tuple ints.
        insert_dims: `False` turns dimensions in `axis` to 1s (use case:
          reduction along `axis` with `keep_dims=True`). `True`, inserts 1s
          according to `axis` (use case: `expand_dims`).
    Returns:
        A tuple of three lists
        - The canonical value for `axis`: always a list, negative values have
          been resolved and values are sorted in ascending order.
        - The output shape: `input_shape` with 1s at the indices in `axis`, for
          use as the `shape` argument of `broadcast_in_dim`.
        - The broadcast dimensions: list of dimensions not in `axis`, for use as
          the `broadcast_dimensions` argument of `broadcast_in_dim`.
    """
    if axis is None:
        raise ValueError("Received `None` value for `axis`")
    if isinstance(axis, int):
        axis = (axis,)
    # Check uniqueness.
    if len(set(axis)) != len(axis):
        raise ValueError(f"Repeated axis in `axis`: {axis}")
    result_dims = len(input_shape)
    if insert_dims:
        result_dims += len(axis)

    # Resolve negative values.
    canonical_axis = []
    for a in axis:
        if not -result_dims <= a < result_dims:
            raise ValueError(
                f"In `axis`, axis {a} is out of bounds for array "
                f"of dimension {result_dims}"
            )
        if a < 0:
            a = a + result_dims
        canonical_axis.append(a)

    # Check uniqueness again after resolving negative values.
    if len(set(canonical_axis)) != len(canonical_axis):
        raise ValueError(f"Repeated axis in `axis`: {canonical_axis}")
    canonical_axis = sorted(canonical_axis)

    # Compute output shape.
    output_shape = list(input_shape)
    for i in canonical_axis:
        if insert_dims:
            output_shape.insert(i, 1)
        else:
            output_shape[i] = 1
    broadcast_dims = [i for i in range(result_dims) if i not in canonical_axis]
    return canonical_axis, output_shape, broadcast_dims


def bcoo_add_indices(x1, x2, sum_duplicates):
    """Add the indices of `x2` to `x1` with zero values.

    Args:
        x1: `BCOO` tensor to add indices to.
        x2: `BCOO` tensor to take the indices to add to x1.
        sum_duplicates: if `True` calls `bcoo_sum_duplicates` on the output.
    Returns:
        a `BCOO` tensor equal to `x1` but with extra zeros at indices in `x2`
        that were missing in `x1`.
    """
    x2_zeros = jnp.zeros(x2.data.shape, x1.data.dtype)
    concat_axis = len(x1.indices.shape) - 2
    output_indices = jnp.concatenate([x1.indices, x2.indices], axis=concat_axis)
    output_data = jnp.concatenate([x1.data, x2_zeros], axis=concat_axis)
    output = jax_sparse.BCOO((output_data, output_indices), shape=x1.shape)
    if sum_duplicates:
        output = jax_sparse.bcoo_sum_duplicates(output)
    return output


def densifying_unary(func):
    """Decorator to add support for `JAXSparse` tensors (including `BCOO`) to a
    non-zero-preserving element-wise unary operator.

    There are requirements on the operator for this decorator to work correctly:

    - The operator must be element-wise
    - The operator must be unary (one input tensor and one output tensor)
    - The operator must return a tensor of the same shape.

    Additional arguments to the function (besides the input tensor) are
    supported. The returned result is a dense tensor.

    Args:
        func: The unary operator to wrap.
    Returns:
        Wrapped function that supports `JAXSparse` tensors.
    """

    @functools.wraps(func)
    def sparse_wrapper(x, *args, **kwargs):
        if isinstance(x, jax_sparse.JAXSparse):
            x = x.todense()
        return func(x, *args, **kwargs)

    return sparse_wrapper


def elementwise_unary(linear):
    """Decorator to add support for `BCOO` sparse tensors to a zero-preserving
    element-wise unary operator.

    There are requirements on the operator for this decorator to work correctly:

    - The operator must be element-wise
    - The operator must be unary (one input tensor and one output tensor)
    - The operator must return a tensor of the same shape, and if it is a
      `BCOO` tensor, the indices of the result must be the same. Therefore:
        - Reduction operations are not supported (e.g. `mean`).
        - Operations for which the result may be dense (e.g. `reciprocal`), or
          the sparse indices depend on the inputs are not supported (e.g.
          `clip`). This implies that `func(0)` must be 0.

    Additional arguments to the function (besides the input tensor) are
    supported as long as they cannot change the indices of the result. For
    instance,`round` is supported, but `clip` is not supported as
    `clip(x, 1.0, 2.0)` would always return a dense tensor.

    Note that if an input sparse tensor contains zero values, the indices and
    the zero values are preserved.

    Args:
        linear: if `True`, means that the operation is such that
            `op(a + b) == op(a) + op(b)`.
    Returns:
        Wrapped function that supports `BCOO` sparse tensors.
    """

    def wrap_elementwise_unary(func):
        @functools.wraps(func)
        def sparse_wrapper(x, *args, **kwargs):
            if isinstance(x, jax_sparse.BCOO):
                if not linear and not x.unique_indices:
                    x = jax_sparse.bcoo_sum_duplicates(x)
                return jax_sparse.BCOO(
                    (func(x.data, *args, **kwargs), x.indices), shape=x.shape
                )
            else:
                return func(x, *args, **kwargs)

        return sparse_wrapper

    return wrap_elementwise_unary


def elementwise_binary_union(linear, use_sparsify):
    """Decorator to add support for `JAXSparse` tensors (including `BCOO`) to an
    element-wise binary operator such that the indices present in the result are
    are the union of the indices in the two operand.

    The primary use case for this is the `add` and `subtract` operators.

    There are requirements on the operator for this decorator to work correctly:

    - The operator must be element-wise.
    - The operator must be binary (two input tensors and one output tensor).
    - Both inputs must be of the same shape or one input must be a scalar.
    - The output must be of the same shape as the (non scalar) inputs.
    - The indices of the output must be the union of the indices of the inputs.
      This implies that func(0, 0) must be 0. As a result, if one operand is
      dense or a scalar, then the result will be dense.

    Additional arguments to the function (besides the input tensors) are not
    supported.

    Note that if the result of the operation is zero at some indices, including
    because the operands were zero at these indices, the zeros and indices are
    preserved.

    The `BCOO` format is the only supported one in all cases. Other formats are
    not supported when `use_sparsify` is `False`.

    Args:
        use_sparsify: indicates that the JAX `sparsify` transform supports this
            operation.
        linear: if `True`, mean that the operation is such that
            `op(a + b, c) == op(a, c) + op(b, c)` and
            `op(a, c + d) == op(a, c) + op(a, d)`.
    Returns:
        Wrapped function that supports `JAXSparse`.
    """

    def wrap_elementwise_binary_union(func):
        sparse_func = jax_sparse.sparsify(func) if use_sparsify else None

        @functools.wraps(func)
        def sparse_wrapper(x1, x2):
            if isinstance(x1, jax_sparse.JAXSparse):
                if isinstance(x2, jax_sparse.JAXSparse):
                    # x1 and x2 are sparse.
                    # The way we use `sparsify` it cannot know that the indices
                    # are the same, so we optimize this case here.
                    if (
                        x1.indices is x2.indices
                        and isinstance(x1, jax_sparse.BCOO)
                        and isinstance(x2, jax_sparse.BCOO)
                    ):
                        if not linear and not x1.unique_indices:
                            x1 = jax_sparse.bcoo_sum_duplicates(x1)
                            x2 = jax_sparse.bcoo_sum_duplicates(x2)
                        return jax_sparse.BCOO(
                            (func(x1.data, x2.data), x1.indices),
                            shape=x1.shape,
                            indices_sorted=x1.indices_sorted,
                            unique_indices=x1.unique_indices,
                        )
                    elif use_sparsify:
                        return sparse_func(x1, x2)
                    elif isinstance(x1, jax_sparse.BCOO) and isinstance(
                        x2, jax_sparse.BCOO
                    ):
                        x1 = bcoo_add_indices(x1, x2, sum_duplicates=not linear)
                        x2 = bcoo_add_indices(x2, x1, sum_duplicates=not linear)
                        return jax_sparse.BCOO(
                            (func(x1.data, x2.data), x1.indices),
                            shape=x1.shape,
                            indices_sorted=True,
                            unique_indices=True,
                        )
                    else:
                        ValueError(
                            "Unsupported sparse format: "
                            f"{x1.__class__} and {x2.__class__}"
                        )
                else:
                    # x1 is sparse, x2 is dense, densify x2.
                    x1 = x1.todense()
            elif isinstance(x2, jax_sparse.JAXSparse):
                # x1 is dense, x2 is sparse, densify x2.
                x2 = x2.todense()
            return func(x1, x2)

        return sparse_wrapper

    return wrap_elementwise_binary_union


def elementwise_division(func):
    """Decorator to add support for `BCOO` sparse tensors to element-wise binary
    division and related operators.

    This decorator is designed for operations related to the division of two
    two operands (e.g. `divide`). It accepts `BCOO` tensors for both the
    dividend and the divisor, but handles them differently based on whether they
    are the dividend or the divisor.

    - If the divisor is sparse, it is densified and the result is dense because
      the result contains Inf or Nan outside of the indices of the dividend.
    - If the dividend is sparse and the divisor is dense, it finds occurrences
      of zeros and NaNs in the divisor. The result may therefore have more
      indices than there were in the dividend to return correct values where the
      divisor was zero or NaN.
    - If the dividend is sparse and the divisor is a scalar, it does the
      division element-wise. Note that the result is incorrectly sparse if the
      scalar divisor is zero.

    Args:
        func: The function to wrap.
    Returns:
        Wrapped function that supports `BCOO` sparse tensors.
    """
    sparse_func = jax_sparse.sparsify(func)

    @functools.wraps(func)
    def sparse_wrapper(x1, x2):
        if isinstance(x1, jax_sparse.JAXSparse):
            if isinstance(x2, jax_sparse.JAXSparse):
                # x1 is sparse and x2 is sparse.
                # Divisor is sparse, meaning we're doing divisions by zero
                # outside of x2.indices, so the result is dense. Densify both.
                x1 = x1.todense()
                x2 = x2.todense()
            elif isinstance(x1, jax_sparse.BCOO):
                if not hasattr(x2, "shape") or len(x2.shape) == 0:
                    # x1 is sparse BCOO, x2 is scalar, apply func element-wise.
                    return jax_sparse.BCOO(
                        (func(x1.data, x2), x1.indices),
                        shape=x1.shape,
                        indices_sorted=x1.indices_sorted,
                        unique_indices=x1.unique_indices,
                    )
                else:
                    # x1 is sparse BCOO, x2 is dense.
                    if not jax_utils.is_in_jax_tracing_scope(x2):
                        # Find zeros and nans in x2 and add indices to x1.
                        # 1. Create a dense mask for zeros and nans.
                        x2_zeros_and_nans = jnp.equal(x2, 0)
                        if not jnp.issubdtype(x2.dtype, jnp.integer):
                            x2_zeros_and_nans = jnp.logical_or(
                                x2_zeros_and_nans, jnp.isnan(x2)
                            )
                        # 2. Make it a BCOO of True values.
                        x2_zeros_and_nans = jax_sparse.bcoo_fromdense(
                            x2_zeros_and_nans,
                            n_batch=x1.n_batch,
                            n_dense=x1.n_dense,
                            index_dtype=x1.indices.dtype,
                        )
                        # 3. Add the indices to x1.
                        x1 = bcoo_add_indices(
                            x1, x2_zeros_and_nans, sum_duplicates=True
                        )
                    return sparse_func(x1, x2)
            else:
                raise ValueError(f"Unsupported sparse format: {x1.__class__}")
        elif isinstance(x2, jax_sparse.JAXSparse):
            # x1 is dense, x2 is sparse, densify x2
            x2 = x2.todense()
        return func(x1, x2)

    return sparse_wrapper



from keras.src.backend.common import global_state


class name_scope:
    """Creates a sub-namespace for variable paths.

    Args:
        name: Name of the current scope (string).
        caller: Optional ID of a caller object (e.g. class instance).
        deduplicate: If `True`, if `caller` was passed,
            and the previous caller matches the current caller,
            and the previous name matches the current name,
            do not reenter a new namespace.
        override_parent: Can be used to provide an absolute path
            which would override any previously opened name scopes.
    """

    def __init__(
        self, name, caller=None, deduplicate=True, override_parent=None
    ):
        if not isinstance(name, str) or "/" in name:
            raise ValueError(
                "Argument `name` must be a string and "
                "cannot contain character `/`. "
                f"Received: name={name}"
            )
        self.name = name
        self.caller = caller
        self.deduplicate = deduplicate
        self.override_parent = override_parent
        if (
            override_parent is None
            and deduplicate
            and getattr(caller, "_parent_path", None) is not None
        ):
            self.override_parent = caller._parent_path
        self._pop_on_exit = False

    def __enter__(self):
        name_scope_stack = global_state.get_global_attribute(
            "name_scope_stack", default=[], set_to_default=True
        )
        if self.deduplicate and name_scope_stack:
            parent_caller = name_scope_stack[-1].caller
            parent_name = name_scope_stack[-1].name
            if (
                self.caller is not None
                and self.caller is parent_caller
                and self.name == parent_name
            ):
                return self
        name_scope_stack.append(self)
        self._pop_on_exit = True
        return self

    def __exit__(self, *args, **kwargs):
        if self._pop_on_exit:
            name_scope_stack = global_state.get_global_attribute(
                "name_scope_stack"
            )
            if name_scope_stack:
                name_scope_stack.pop()


def current_path():
    name_scope_stack = global_state.get_global_attribute("name_scope_stack")
    if name_scope_stack is None:
        return ""
    parts = []
    for entry in name_scope_stack:
        if entry.override_parent is not None:
            parts = [p for p in entry.override_parent.split("/") if p]
        parts.append(entry.name)
    return "/".join(parts)



import itertools

import numpy as np

from keras.src import tree
from keras.src.trainers.data_adapters import data_adapter_utils
from keras.src.trainers.data_adapters.data_adapter import DataAdapter
from keras.src.utils.module_utils import grain
from keras.src.utils.module_utils import tensorflow as tf


class GrainDatasetAdapter(DataAdapter):
    """Adapter that handles `grain.DataLoader`, `grain.MapDataset` and
    `grain.IterDataset`.
    """

    def __init__(self, dataset):
        """Initialize the GrainDatasetAdapter.

        Args:
            dataset: A Grain dataset instance. Must be one of
                `grain.DataLoader`, `grain.MapDataset`, or `grain.IterDataset`.
        """

        if not isinstance(
            dataset, (grain.MapDataset, grain.IterDataset, grain.DataLoader)
        ):
            raise ValueError(
                "Expected `dataset` to be a grain.MapDataset, "
                "grain.IterDataset or grain.DataLoader. "
                f"Received: {dataset} of type {type(dataset)}"
            )

        self._dataset = dataset

        batch_size, output_signature = self._get_dataset_info(dataset)
        self._batch_size = batch_size
        self._output_signature = output_signature
        self._output_tf_signature = None

    def _get_dataset_info(self, dataset):
        """Get the `batch_size` and `output_signature` from the dataset.

        We use a small list of batches to infer the `batch_size` and
        `output_signature`.
        """
        batches = list(
            itertools.islice(
                dataset, data_adapter_utils.NUM_BATCHES_FOR_TENSOR_SPEC
            )
        )
        output_signature = data_adapter_utils.get_keras_tensor_spec(batches)
        flat_output_signature = tree.flatten(output_signature)
        batch_size = flat_output_signature[0].shape[0]
        if batch_size is not None:
            batch_size = int(batch_size)
        return batch_size, output_signature

    def get_numpy_iterator(self):
        # Workaround for internal change in Grain which isn't a part of a
        # release yet.
        # TODO(abheesht17): Remove this after the next Grain release.
        try:
            from grain._src.python.shared_memory_array import (
                SharedMemoryArrayMetadata,
            )
        except ImportError:
            from grain._src.python.ipc.shared_memory_array import (
                SharedMemoryArrayMetadata,
            )

        def convert_to_numpy(x):
            if isinstance(x, (np.ndarray, SharedMemoryArrayMetadata)):
                return x
            else:
                # Using `__array__` should handle `tf.Tensor`, `jax.np.ndarray`,
                # `torch.Tensor`, as well as any other tensor-like object that
                # has added numpy support.
                if hasattr(x, "__array__"):
                    if data_adapter_utils.is_torch_tensor(x):
                        x = x.cpu()
                    x = np.asarray(x)
                return x

        class ConvertToNumpy(grain.transforms.Map):
            def map(self, x):
                return tree.map_structure(
                    convert_to_numpy, x, none_is_leaf=False
                )

        if isinstance(self._dataset, (grain.MapDataset, grain.IterDataset)):
            dataset = self._dataset.map(ConvertToNumpy())
        else:
            # Instantiate a new `DataLoader`.
            dataset = grain.DataLoader(
                data_source=self._dataset._data_source,
                sampler=self._dataset._sampler,
                # Append `ConvertToNumpy`.
                operations=list(self._dataset._operations) + [ConvertToNumpy()],
                worker_count=self._dataset._multiprocessing_options.num_workers,
                worker_buffer_size=self._dataset._multiprocessing_options.per_worker_buffer_size,
                shard_options=self._dataset._shard_options,
                read_options=self._dataset._read_options,
                enable_profiling=self._dataset._multiprocessing_options.enable_profiling,
            )
        return dataset

    def get_jax_iterator(self):
        def convert_to_jax_compatible(x):
            if data_adapter_utils.is_scipy_sparse(x):
                x = data_adapter_utils.scipy_sparse_to_jax_sparse(x)
            elif data_adapter_utils.is_tensorflow_sparse(x):
                x = data_adapter_utils.tf_sparse_to_jax_sparse(x)
            return x

        class ConvertToJaxCompatible(grain.transforms.Map):
            def map(self, x):
                return tree.map_structure(
                    convert_to_jax_compatible, x, none_is_leaf=False
                )

        if isinstance(self._dataset, (grain.MapDataset, grain.IterDataset)):
            dataset = self._dataset.map(ConvertToJaxCompatible())
        else:
            # Instantiate a new `DataLoader`.
            dataset = grain.DataLoader(
                data_source=self._dataset._data_source,
                sampler=self._dataset._sampler,
                # Append `ConvertToJaxCompatible`.
                operations=list(self._dataset._operations)
                + [ConvertToJaxCompatible()],
                worker_count=self._dataset._multiprocessing_options.num_workers,
                worker_buffer_size=self._dataset._multiprocessing_options.per_worker_buffer_size,
                shard_options=self._dataset._shard_options,
                read_options=self._dataset._read_options,
                enable_profiling=self._dataset._multiprocessing_options.enable_profiling,
            )
        return dataset

    def get_tf_dataset(self):
        def convert_to_tf(x):
            if x is None:
                return tf.experimental.Optional.empty(None)
            if data_adapter_utils.is_scipy_sparse(x):
                x = data_adapter_utils.scipy_sparse_to_tf_sparse(x)
            elif data_adapter_utils.is_jax_sparse(x):
                x = data_adapter_utils.jax_sparse_to_tf_sparse(x)
            return x

        class ConvertToTF(grain.transforms.Map):
            def map(self, x):
                return tree.map_structure(convert_to_tf, x)

        # `tf.data.Dataset.from_generator` does not support lists as output.
        # We convert lists to tuples.
        class ListToTuple(grain.transforms.Map):
            def map(self, x):
                return tree.lists_to_tuples(x)

        if isinstance(self._dataset, (grain.MapDataset, grain.IterDataset)):
            dataset = self._dataset.map(ConvertToTF())
            dataset = dataset.map(ListToTuple())
        else:
            # Instantiate a new `DataLoader`.
            dataset = grain.DataLoader(
                data_source=self._dataset._data_source,
                sampler=self._dataset._sampler,
                # Append `ConvertToTF` and `ListToTuple`.
                operations=list(self._dataset._operations)
                + [ConvertToTF(), ListToTuple()],
                worker_count=self._dataset._multiprocessing_options.num_workers,
                worker_buffer_size=self._dataset._multiprocessing_options.per_worker_buffer_size,
                shard_options=self._dataset._shard_options,
                read_options=self._dataset._read_options,
                enable_profiling=self._dataset._multiprocessing_options.enable_profiling,
            )

        if self._output_tf_signature is None:
            self._output_tf_signature = tree.map_structure(
                data_adapter_utils.convert_to_tf_tensor_spec,
                self._output_signature,
            )

        return tf.data.Dataset.from_generator(
            lambda: dataset, output_signature=self._output_tf_signature
        )

    def get_torch_dataloader(self):
        import torch.utils.data as torch_data

        class ConverterIterableDataset(torch_data.IterableDataset):
            def __init__(self, iterable):
                super().__init__()
                self.iterable = iterable

            def __iter__(self):
                return iter(self.iterable)

        # `batch_size=None` indicates that we should not re-batch
        return torch_data.DataLoader(
            ConverterIterableDataset(self._dataset), batch_size=None
        )

    @property
    def num_batches(self):
        return None

    @property
    def batch_size(self):
        return self._batch_size

    @property
    def has_partial_batch(self):
        return None

    @property
    def partial_batch_size(self):
        return None



from keras.src.api_export import keras_export
from keras.src.layers.preprocessing.image_preprocessing.base_image_preprocessing_layer import (  # noqa: E501
    BaseImagePreprocessingLayer,
)
from keras.src.layers.preprocessing.image_preprocessing.base_image_preprocessing_layer import (  # noqa: E501
    base_image_preprocessing_transform_example,
)
from keras.src.layers.preprocessing.image_preprocessing.bounding_boxes.converters import (  # noqa: E501
    clip_to_image_size,
)
from keras.src.layers.preprocessing.image_preprocessing.bounding_boxes.converters import (  # noqa: E501
    convert_format,
)
from keras.src.random.seed_generator import SeedGenerator
from keras.src.utils import backend_utils


@keras_export("keras.layers.RandomPerspective")
class RandomPerspective(BaseImagePreprocessingLayer):
    """A preprocessing layer that applies random perspective transformations.

    This layer distorts the perspective of input images by shifting their
    corner points, simulating a 3D-like transformation. The amount of distortion
    is controlled by the `factor` and `scale` parameters.

    **Note:** This layer is safe to use inside a `tf.data` or `grain` pipeline
    (independently of which backend you're using).

    Args:
        factor: A float or a tuple of two floats.
            Represents the probability of applying the perspective
            transformation to each image in the batch.
            - `factor=0.0` ensures no transformation is applied.
            - `factor=1.0` means the transformation is always applied.
            - If a tuple `(min, max)` is provided, a probability is randomly
              sampled between `min` and `max` for each image.
            - If a single float is given, the probability is sampled between
              `0.0` and the provided float.
            Default is 1.0.
        scale: A float defining the relative amount of perspective shift.
            Determines how much the image corners are displaced, affecting
            the intensity of the perspective effect.
        interpolation: Interpolation mode. Supported values: `"nearest"`,
            `"bilinear"`.
        fill_value: a float represents the value to be filled outside the
            boundaries when `fill_mode="constant"`.
        seed: Integer. Used to create a random seed.

    Example:

    {{base_image_preprocessing_transform_example}}
    """

    _USE_BASE_FACTOR = False
    _FACTOR_BOUNDS = (0, 1)
    _SUPPORTED_INTERPOLATION = ("nearest", "bilinear")

    def __init__(
        self,
        factor=1.0,
        scale=1.0,
        interpolation="bilinear",
        fill_value=0.0,
        seed=None,
        data_format=None,
        **kwargs,
    ):
        super().__init__(data_format=data_format, **kwargs)
        self._set_factor(factor)
        self.scale = scale
        self.fill_value = fill_value
        self.interpolation = interpolation
        self.seed = seed
        self.generator = SeedGenerator(seed)
        self.supports_jit = False

        if scale < 0.0 or scale > 1.0:
            raise ValueError(
                "The `scale` argument should be a number "
                "in the range "
                f"[0,1]. "
                f"Received: scale={scale}"
            )

        if interpolation not in self._SUPPORTED_INTERPOLATION:
            raise NotImplementedError(
                f"Unknown `interpolation` {interpolation}. Expected of one "
                f"{self._SUPPORTED_INTERPOLATION}."
            )

        if self.data_format == "channels_first":
            self.height_axis = -2
            self.width_axis = -1
            self.channel_axis = -3
        else:
            self.height_axis = -3
            self.width_axis = -2
            self.channel_axis = -1

    def get_random_transformation(self, data, training=True, seed=None):
        if not training:
            return None

        if isinstance(data, dict):
            images = data["images"]
        else:
            images = data

        images_shape = self.backend.shape(images)
        unbatched = len(images_shape) == 3
        if unbatched:
            batch_size = 1
        else:
            batch_size = images_shape[0]
        height, width = (
            images.shape[self.height_axis],
            images.shape[self.width_axis],
        )

        seed = seed or self._get_seed_generator(self.backend._backend)

        transformation_probability = self.backend.random.uniform(
            shape=(batch_size,),
            minval=self.factor[0],
            maxval=self.factor[1],
            seed=seed,
        )

        random_threshold = self.backend.random.uniform(
            shape=(batch_size,),
            minval=0.0,
            maxval=1.0,
            seed=seed,
        )
        apply_perspective = random_threshold < transformation_probability

        perspective_factor = self.backend.random.uniform(
            shape=(batch_size, 4, 2),
            minval=-0.5 * self.scale,
            maxval=0.5 * self.scale,
            seed=seed,
            dtype=self.compute_dtype,
        )

        start_points = self.backend.convert_to_tensor(
            [
                [
                    [0.0, 0.0],
                    [width - 1, 0.0],
                    [0.0, height - 1],
                    [width - 1, height - 1],
                ]
            ],
            dtype=self.compute_dtype,
        )

        start_points = self.backend.numpy.repeat(
            start_points, batch_size, axis=0
        )
        end_points = start_points + start_points * perspective_factor

        return {
            "apply_perspective": apply_perspective,
            "start_points": start_points,
            "end_points": end_points,
            "input_shape": images_shape,
        }

    def _transform_images(self, images, transformation, interpolation):
        if transformation is None:
            return images

        inputs_shape = self.backend.shape(images)
        unbatched = len(inputs_shape) == 3
        if unbatched:
            images = self.backend.numpy.expand_dims(images, axis=0)

        start_points = transformation["start_points"]
        end_points = transformation["end_points"]

        outputs = self.backend.image.perspective_transform(
            images,
            start_points,
            end_points,
            interpolation=interpolation,
            fill_value=self.fill_value,
            data_format=self.data_format,
        )

        apply_perspective = transformation["apply_perspective"]
        outputs = self.backend.numpy.where(
            apply_perspective[:, None, None, None],
            outputs,
            images,
        )

        if unbatched:
            outputs = self.backend.numpy.squeeze(outputs, axis=0)
        return outputs

    def transform_bounding_boxes(
        self,
        bounding_boxes,
        transformation,
        training=True,
    ):
        if training and transformation is not None:
            if backend_utils.in_tf_graph():
                self.backend.set_backend("tensorflow")

            input_height, input_width = (
                transformation["input_shape"][self.height_axis],
                transformation["input_shape"][self.width_axis],
            )

            bounding_boxes = convert_format(
                bounding_boxes,
                source=self.bounding_box_format,
                target="xyxy",
                height=input_height,
                width=input_width,
            )

            boxes = bounding_boxes["boxes"]
            x0, y0, x1, y1 = self.backend.numpy.split(boxes, 4, axis=-1)

            start_points = transformation["start_points"]
            end_points = transformation["end_points"]
            transform = self.backend.image.compute_homography_matrix(
                start_points, end_points
            )
            transform = self.backend.numpy.expand_dims(transform, axis=1)
            transform = self.backend.cast(transform, dtype=self.compute_dtype)

            corners = [
                self._get_transformed_coordinates(x, y, transform)
                for x, y in [(x0, y0), (x1, y1), (x0, y1), (x1, y0)]
            ]
            x_corners, y_corners = zip(*corners)

            xs = self.backend.numpy.stack(x_corners, axis=-1)
            ys = self.backend.numpy.stack(y_corners, axis=-1)

            min_x, max_x = (
                self.backend.numpy.min(xs, axis=-1),
                self.backend.numpy.max(xs, axis=-1),
            )
            min_y, max_y = (
                self.backend.numpy.min(ys, axis=-1),
                self.backend.numpy.max(ys, axis=-1),
            )

            min_x = self.backend.numpy.expand_dims(min_x, axis=-1)
            max_x = self.backend.numpy.expand_dims(max_x, axis=-1)
            min_y = self.backend.numpy.expand_dims(min_y, axis=-1)
            max_y = self.backend.numpy.expand_dims(max_y, axis=-1)

            boxes = self.backend.numpy.concatenate(
                [min_x, min_y, max_x, max_y], axis=-1
            )

            apply_perspective = self.backend.core.convert_to_tensor(
                transformation["apply_perspective"], dtype=boxes.dtype
            )

            bounding_boxes["boxes"] = self.backend.numpy.where(
                apply_perspective[:, None, None],
                boxes,
                bounding_boxes["boxes"],
            )

            bounding_boxes = clip_to_image_size(
                bounding_boxes=bounding_boxes,
                height=input_height,
                width=input_width,
                bounding_box_format="xyxy",
            )

            self.backend.reset()

        return bounding_boxes

    def _get_transformed_coordinates(
        self, x_coords, y_coords, transformation_matrix
    ):
        backend = self.backend

        batch_size = backend.shape(transformation_matrix)[0]

        homogeneous_transform = backend.numpy.concatenate(
            [transformation_matrix, backend.numpy.ones((batch_size, 1, 1))],
            axis=-1,
        )
        homogeneous_transform = backend.numpy.reshape(
            homogeneous_transform, (batch_size, 3, 3)
        )

        inverse_transform = backend.linalg.inv(homogeneous_transform)

        ones_column = backend.numpy.ones_like(x_coords)
        homogeneous_coords = backend.numpy.concatenate(
            [x_coords, y_coords, ones_column], axis=-1
        )

        homogeneous_coords = backend.numpy.moveaxis(homogeneous_coords, -1, -2)
        transformed_coords = backend.numpy.matmul(
            inverse_transform, homogeneous_coords
        )
        transformed_coords = backend.numpy.moveaxis(transformed_coords, -1, -2)

        x_transformed = transformed_coords[..., 0] / transformed_coords[..., 2]
        y_transformed = transformed_coords[..., 1] / transformed_coords[..., 2]

        return x_transformed, y_transformed

    def transform_labels(self, labels, transformation, training=True):
        return labels

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        base_config = super().get_config()
        config = {
            "factor": self.factor,
            "scale": self.scale,
            "interpolation": self.interpolation,
            "fill_value": self.fill_value,
            "seed": self.seed,
        }
        return {**base_config, **config}


if RandomPerspective.__doc__ is not None:
    RandomPerspective.__doc__ = RandomPerspective.__doc__.replace(
        "{{base_image_preprocessing_transform_example}}",
        base_image_preprocessing_transform_example.replace(
            "{LayerName}", "RandomPerspective"
        ),
    )



from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.dtype_policies import dtype_policy
from keras.src.dtype_policies.dtype_policy import QUANTIZATION_MODES
from keras.src.dtype_policies.dtype_policy import AWQDTypePolicy
from keras.src.dtype_policies.dtype_policy import DTypePolicy
from keras.src.dtype_policies.dtype_policy import FloatDTypePolicy
from keras.src.dtype_policies.dtype_policy import GPTQDTypePolicy
from keras.src.dtype_policies.dtype_policy import Int4DTypePolicy
from keras.src.dtype_policies.dtype_policy import QuantizedDTypePolicy
from keras.src.dtype_policies.dtype_policy import QuantizedFloat8DTypePolicy
from keras.src.dtype_policies.dtype_policy_map import DTypePolicyMap

ALL_OBJECTS = {
    AWQDTypePolicy,
    DTypePolicy,
    FloatDTypePolicy,
    QuantizedDTypePolicy,
    QuantizedFloat8DTypePolicy,
    DTypePolicyMap,
    GPTQDTypePolicy,
    Int4DTypePolicy,
}
ALL_OBJECTS_DICT = {cls.__name__: cls for cls in ALL_OBJECTS}


@keras_export("keras.dtype_policies.serialize")
def serialize(dtype_policy):
    """Serializes `DTypePolicy` instance.

    Args:
        dtype_policy: A Keras `DTypePolicy` instance.

    Returns:
        `DTypePolicy` configuration dictionary.
    """
    from keras.src.saving import serialization_lib

    return serialization_lib.serialize_keras_object(dtype_policy)


@keras_export("keras.dtype_policies.deserialize")
def deserialize(config, custom_objects=None):
    """Deserializes a serialized `DTypePolicy` instance.

    Args:
        config: `DTypePolicy` configuration.
        custom_objects: Optional dictionary mapping names (strings) to custom
            objects (classes and functions) to be considered during
            deserialization.

    Returns:
        A Keras `DTypePolicy` instance.
    """
    from keras.src.saving import serialization_lib

    return serialization_lib.deserialize_keras_object(
        config,
        module_objects=ALL_OBJECTS_DICT,
        custom_objects=custom_objects,
    )


@keras_export("keras.dtype_policies.get")
def get(identifier):
    """Retrieves a Keras `DTypePolicy` instance.

    The `identifier` may be the string name of a `DTypePolicy` class.

    >>> policy = dtype_policies.get("mixed_bfloat16")
    >>> type(policy)
    <class '...DTypePolicy'>

    You can also specify `config` of the dtype policy to this function by
    passing dict containing `class_name` and `config` as an identifier. Also
    note that the `class_name` must map to a `DTypePolicy` class

    >>> identifier = {"class_name": "DTypePolicy",
    ...               "config": {"name": "float32"}}
    >>> policy = dtype_policies.get(identifier)
    >>> type(policy)
    <class '...DTypePolicy'>

    Args:
        identifier: A dtype policy identifier. One of `None` or string name of a
            `DTypePolicy` or `DTypePolicy` configuration dictionary or a
            `DTypePolicy` instance.

    Returns:
        A Keras `DTypePolicy` instance.
    """
    from keras.src.dtype_policies.dtype_policy import (
        _get_quantized_dtype_policy_by_str,
    )

    if identifier is None:
        return dtype_policy.dtype_policy()
    if isinstance(identifier, DTypePolicy):
        return identifier
    if isinstance(identifier, dict):
        return deserialize(identifier)
    if isinstance(identifier, str):
        if identifier.startswith(QUANTIZATION_MODES):
            return _get_quantized_dtype_policy_by_str(identifier)
        else:
            return DTypePolicy(identifier)
    try:
        return DTypePolicy(backend.standardize_dtype(identifier))
    except:
        raise ValueError(
            "Cannot interpret `dtype` argument. Expected a string "
            f"or an instance of DTypePolicy. Received: dtype={identifier}"
        )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.ops import image as image
from keras.ops import linalg as linalg
from keras.ops import nn as nn
from keras.ops import numpy as numpy
from keras.src.ops.core import associative_scan as associative_scan
from keras.src.ops.core import cast as cast
from keras.src.ops.core import cond as cond
from keras.src.ops.core import convert_to_numpy as convert_to_numpy
from keras.src.ops.core import convert_to_tensor as convert_to_tensor
from keras.src.ops.core import custom_gradient as custom_gradient
from keras.src.ops.core import dtype as dtype
from keras.src.ops.core import fori_loop as fori_loop
from keras.src.ops.core import is_tensor as is_tensor
from keras.src.ops.core import map as map
from keras.src.ops.core import saturate_cast as saturate_cast
from keras.src.ops.core import scan as scan
from keras.src.ops.core import scatter as scatter
from keras.src.ops.core import scatter_update as scatter_update
from keras.src.ops.core import shape as shape
from keras.src.ops.core import slice as slice
from keras.src.ops.core import slice_update as slice_update
from keras.src.ops.core import stop_gradient as stop_gradient
from keras.src.ops.core import switch as switch
from keras.src.ops.core import unstack as unstack
from keras.src.ops.core import vectorized_map as vectorized_map
from keras.src.ops.core import while_loop as while_loop
from keras.src.ops.einops import rearrange as rearrange
from keras.src.ops.linalg import cholesky as cholesky
from keras.src.ops.linalg import cholesky_inverse as cholesky_inverse
from keras.src.ops.linalg import det as det
from keras.src.ops.linalg import eig as eig
from keras.src.ops.linalg import eigh as eigh
from keras.src.ops.linalg import inv as inv
from keras.src.ops.linalg import jvp as jvp
from keras.src.ops.linalg import lstsq as lstsq
from keras.src.ops.linalg import lu_factor as lu_factor
from keras.src.ops.linalg import norm as norm
from keras.src.ops.linalg import qr as qr
from keras.src.ops.linalg import solve as solve
from keras.src.ops.linalg import solve_triangular as solve_triangular
from keras.src.ops.linalg import svd as svd
from keras.src.ops.math import erf as erf
from keras.src.ops.math import erfinv as erfinv
from keras.src.ops.math import extract_sequences as extract_sequences
from keras.src.ops.math import fft as fft
from keras.src.ops.math import fft2 as fft2
from keras.src.ops.math import ifft2 as ifft2
from keras.src.ops.math import in_top_k as in_top_k
from keras.src.ops.math import irfft as irfft
from keras.src.ops.math import istft as istft
from keras.src.ops.math import logdet as logdet
from keras.src.ops.math import logsumexp as logsumexp
from keras.src.ops.math import rfft as rfft
from keras.src.ops.math import rsqrt as rsqrt
from keras.src.ops.math import segment_max as segment_max
from keras.src.ops.math import segment_sum as segment_sum
from keras.src.ops.math import stft as stft
from keras.src.ops.math import top_k as top_k
from keras.src.ops.math import view_as_complex as view_as_complex
from keras.src.ops.math import view_as_real as view_as_real
from keras.src.ops.nn import adaptive_average_pool as adaptive_average_pool
from keras.src.ops.nn import adaptive_max_pool as adaptive_max_pool
from keras.src.ops.nn import average_pool as average_pool
from keras.src.ops.nn import batch_normalization as batch_normalization
from keras.src.ops.nn import binary_crossentropy as binary_crossentropy
from keras.src.ops.nn import (
    categorical_crossentropy as categorical_crossentropy,
)
from keras.src.ops.nn import celu as celu
from keras.src.ops.nn import conv as conv
from keras.src.ops.nn import conv_transpose as conv_transpose
from keras.src.ops.nn import ctc_decode as ctc_decode
from keras.src.ops.nn import ctc_loss as ctc_loss
from keras.src.ops.nn import depth_to_space as depth_to_space
from keras.src.ops.nn import depthwise_conv as depthwise_conv
from keras.src.ops.nn import dot_product_attention as dot_product_attention
from keras.src.ops.nn import elu as elu
from keras.src.ops.nn import fold as fold
from keras.src.ops.nn import gelu as gelu
from keras.src.ops.nn import glu as glu
from keras.src.ops.nn import hard_shrink as hard_shrink
from keras.src.ops.nn import hard_sigmoid as hard_sigmoid
from keras.src.ops.nn import hard_silu as hard_silu
from keras.src.ops.nn import hard_silu as hard_swish
from keras.src.ops.nn import hard_tanh as hard_tanh
from keras.src.ops.nn import layer_normalization as layer_normalization
from keras.src.ops.nn import leaky_relu as leaky_relu
from keras.src.ops.nn import log_sigmoid as log_sigmoid
from keras.src.ops.nn import log_softmax as log_softmax
from keras.src.ops.nn import max_pool as max_pool
from keras.src.ops.nn import moments as moments
from keras.src.ops.nn import multi_hot as multi_hot
from keras.src.ops.nn import normalize as normalize
from keras.src.ops.nn import one_hot as one_hot
from keras.src.ops.nn import polar as polar
from keras.src.ops.nn import psnr as psnr
from keras.src.ops.nn import relu as relu
from keras.src.ops.nn import relu6 as relu6
from keras.src.ops.nn import rms_normalization as rms_normalization
from keras.src.ops.nn import selu as selu
from keras.src.ops.nn import separable_conv as separable_conv
from keras.src.ops.nn import sigmoid as sigmoid
from keras.src.ops.nn import silu as silu
from keras.src.ops.nn import silu as swish
from keras.src.ops.nn import soft_shrink as soft_shrink
from keras.src.ops.nn import softmax as softmax
from keras.src.ops.nn import softplus as softplus
from keras.src.ops.nn import softsign as softsign
from keras.src.ops.nn import space_to_depth as space_to_depth
from keras.src.ops.nn import (
    sparse_categorical_crossentropy as sparse_categorical_crossentropy,
)
from keras.src.ops.nn import sparse_plus as sparse_plus
from keras.src.ops.nn import sparse_sigmoid as sparse_sigmoid
from keras.src.ops.nn import sparsemax as sparsemax
from keras.src.ops.nn import squareplus as squareplus
from keras.src.ops.nn import tanh_shrink as tanh_shrink
from keras.src.ops.nn import threshold as threshold
from keras.src.ops.nn import unfold as unfold
from keras.src.ops.numpy import abs as abs
from keras.src.ops.numpy import absolute as absolute
from keras.src.ops.numpy import add as add
from keras.src.ops.numpy import all as all
from keras.src.ops.numpy import allclose as allclose
from keras.src.ops.numpy import amax as amax
from keras.src.ops.numpy import amin as amin
from keras.src.ops.numpy import angle as angle
from keras.src.ops.numpy import any as any
from keras.src.ops.numpy import append as append
from keras.src.ops.numpy import arange as arange
from keras.src.ops.numpy import arccos as arccos
from keras.src.ops.numpy import arccosh as arccosh
from keras.src.ops.numpy import arcsin as arcsin
from keras.src.ops.numpy import arcsinh as arcsinh
from keras.src.ops.numpy import arctan as arctan
from keras.src.ops.numpy import arctan2 as arctan2
from keras.src.ops.numpy import arctanh as arctanh
from keras.src.ops.numpy import argmax as argmax
from keras.src.ops.numpy import argmin as argmin
from keras.src.ops.numpy import argpartition as argpartition
from keras.src.ops.numpy import argsort as argsort
from keras.src.ops.numpy import array as array
from keras.src.ops.numpy import array_split as array_split
from keras.src.ops.numpy import average as average
from keras.src.ops.numpy import bartlett as bartlett
from keras.src.ops.numpy import bincount as bincount
from keras.src.ops.numpy import bitwise_and as bitwise_and
from keras.src.ops.numpy import bitwise_invert as bitwise_invert
from keras.src.ops.numpy import bitwise_left_shift as bitwise_left_shift
from keras.src.ops.numpy import bitwise_not as bitwise_not
from keras.src.ops.numpy import bitwise_or as bitwise_or
from keras.src.ops.numpy import bitwise_right_shift as bitwise_right_shift
from keras.src.ops.numpy import bitwise_xor as bitwise_xor
from keras.src.ops.numpy import blackman as blackman
from keras.src.ops.numpy import broadcast_to as broadcast_to
from keras.src.ops.numpy import cbrt as cbrt
from keras.src.ops.numpy import ceil as ceil
from keras.src.ops.numpy import clip as clip
from keras.src.ops.numpy import concatenate as concatenate
from keras.src.ops.numpy import conj as conj
from keras.src.ops.numpy import conjugate as conjugate
from keras.src.ops.numpy import copy as copy
from keras.src.ops.numpy import corrcoef as corrcoef
from keras.src.ops.numpy import correlate as correlate
from keras.src.ops.numpy import cos as cos
from keras.src.ops.numpy import cosh as cosh
from keras.src.ops.numpy import count_nonzero as count_nonzero
from keras.src.ops.numpy import cross as cross
from keras.src.ops.numpy import cumprod as cumprod
from keras.src.ops.numpy import cumsum as cumsum
from keras.src.ops.numpy import deg2rad as deg2rad
from keras.src.ops.numpy import diag as diag
from keras.src.ops.numpy import diagflat as diagflat
from keras.src.ops.numpy import diagonal as diagonal
from keras.src.ops.numpy import diff as diff
from keras.src.ops.numpy import digitize as digitize
from keras.src.ops.numpy import divide as divide
from keras.src.ops.numpy import divide_no_nan as divide_no_nan
from keras.src.ops.numpy import dot as dot
from keras.src.ops.numpy import dstack as dstack
from keras.src.ops.numpy import einsum as einsum
from keras.src.ops.numpy import empty as empty
from keras.src.ops.numpy import empty_like as empty_like
from keras.src.ops.numpy import equal as equal
from keras.src.ops.numpy import exp as exp
from keras.src.ops.numpy import exp2 as exp2
from keras.src.ops.numpy import expand_dims as expand_dims
from keras.src.ops.numpy import expm1 as expm1
from keras.src.ops.numpy import eye as eye
from keras.src.ops.numpy import flip as flip
from keras.src.ops.numpy import fliplr as fliplr
from keras.src.ops.numpy import flipud as flipud
from keras.src.ops.numpy import floor as floor
from keras.src.ops.numpy import floor_divide as floor_divide
from keras.src.ops.numpy import fmod as fmod
from keras.src.ops.numpy import full as full
from keras.src.ops.numpy import full_like as full_like
from keras.src.ops.numpy import gcd as gcd
from keras.src.ops.numpy import geomspace as geomspace
from keras.src.ops.numpy import get_item as get_item
from keras.src.ops.numpy import greater as greater
from keras.src.ops.numpy import greater_equal as greater_equal
from keras.src.ops.numpy import hamming as hamming
from keras.src.ops.numpy import hanning as hanning
from keras.src.ops.numpy import heaviside as heaviside
from keras.src.ops.numpy import histogram as histogram
from keras.src.ops.numpy import hsplit as hsplit
from keras.src.ops.numpy import hstack as hstack
from keras.src.ops.numpy import hypot as hypot
from keras.src.ops.numpy import i0 as i0
from keras.src.ops.numpy import identity as identity
from keras.src.ops.numpy import imag as imag
from keras.src.ops.numpy import inner as inner
from keras.src.ops.numpy import isclose as isclose
from keras.src.ops.numpy import isfinite as isfinite
from keras.src.ops.numpy import isin as isin
from keras.src.ops.numpy import isinf as isinf
from keras.src.ops.numpy import isnan as isnan
from keras.src.ops.numpy import isneginf as isneginf
from keras.src.ops.numpy import isposinf as isposinf
from keras.src.ops.numpy import isreal as isreal
from keras.src.ops.numpy import kaiser as kaiser
from keras.src.ops.numpy import kron as kron
from keras.src.ops.numpy import lcm as lcm
from keras.src.ops.numpy import ldexp as ldexp
from keras.src.ops.numpy import left_shift as left_shift
from keras.src.ops.numpy import less as less
from keras.src.ops.numpy import less_equal as less_equal
from keras.src.ops.numpy import linspace as linspace
from keras.src.ops.numpy import log as log
from keras.src.ops.numpy import log1p as log1p
from keras.src.ops.numpy import log2 as log2
from keras.src.ops.numpy import log10 as log10
from keras.src.ops.numpy import logaddexp as logaddexp
from keras.src.ops.numpy import logaddexp2 as logaddexp2
from keras.src.ops.numpy import logical_and as logical_and
from keras.src.ops.numpy import logical_not as logical_not
from keras.src.ops.numpy import logical_or as logical_or
from keras.src.ops.numpy import logical_xor as logical_xor
from keras.src.ops.numpy import logspace as logspace
from keras.src.ops.numpy import matmul as matmul
from keras.src.ops.numpy import max as max
from keras.src.ops.numpy import maximum as maximum
from keras.src.ops.numpy import mean as mean
from keras.src.ops.numpy import median as median
from keras.src.ops.numpy import meshgrid as meshgrid
from keras.src.ops.numpy import min as min
from keras.src.ops.numpy import minimum as minimum
from keras.src.ops.numpy import mod as mod
from keras.src.ops.numpy import moveaxis as moveaxis
from keras.src.ops.numpy import multiply as multiply
from keras.src.ops.numpy import nan_to_num as nan_to_num
from keras.src.ops.numpy import nanargmax as nanargmax
from keras.src.ops.numpy import nanargmin as nanargmin
from keras.src.ops.numpy import nancumprod as nancumprod
from keras.src.ops.numpy import nancumsum as nancumsum
from keras.src.ops.numpy import nanmax as nanmax
from keras.src.ops.numpy import nanmean as nanmean
from keras.src.ops.numpy import nanmedian as nanmedian
from keras.src.ops.numpy import nanmin as nanmin
from keras.src.ops.numpy import nanprod as nanprod
from keras.src.ops.numpy import nanquantile as nanquantile
from keras.src.ops.numpy import nanstd as nanstd
from keras.src.ops.numpy import nansum as nansum
from keras.src.ops.numpy import nanvar as nanvar
from keras.src.ops.numpy import ndim as ndim
from keras.src.ops.numpy import negative as negative
from keras.src.ops.numpy import nextafter as nextafter
from keras.src.ops.numpy import nonzero as nonzero
from keras.src.ops.numpy import not_equal as not_equal
from keras.src.ops.numpy import ones as ones
from keras.src.ops.numpy import ones_like as ones_like
from keras.src.ops.numpy import outer as outer
from keras.src.ops.numpy import pad as pad
from keras.src.ops.numpy import power as power
from keras.src.ops.numpy import prod as prod
from keras.src.ops.numpy import ptp as ptp
from keras.src.ops.numpy import quantile as quantile
from keras.src.ops.numpy import rad2deg as rad2deg
from keras.src.ops.numpy import ravel as ravel
from keras.src.ops.numpy import real as real
from keras.src.ops.numpy import reciprocal as reciprocal
from keras.src.ops.numpy import repeat as repeat
from keras.src.ops.numpy import reshape as reshape
from keras.src.ops.numpy import right_shift as right_shift
from keras.src.ops.numpy import roll as roll
from keras.src.ops.numpy import rot90 as rot90
from keras.src.ops.numpy import round as round
from keras.src.ops.numpy import searchsorted as searchsorted
from keras.src.ops.numpy import select as select
from keras.src.ops.numpy import sign as sign
from keras.src.ops.numpy import signbit as signbit
from keras.src.ops.numpy import sin as sin
from keras.src.ops.numpy import sinc as sinc
from keras.src.ops.numpy import sinh as sinh
from keras.src.ops.numpy import size as size
from keras.src.ops.numpy import slogdet as slogdet
from keras.src.ops.numpy import sort as sort
from keras.src.ops.numpy import split as split
from keras.src.ops.numpy import sqrt as sqrt
from keras.src.ops.numpy import square as square
from keras.src.ops.numpy import squeeze as squeeze
from keras.src.ops.numpy import stack as stack
from keras.src.ops.numpy import std as std
from keras.src.ops.numpy import subtract as subtract
from keras.src.ops.numpy import sum as sum
from keras.src.ops.numpy import swapaxes as swapaxes
from keras.src.ops.numpy import take as take
from keras.src.ops.numpy import take_along_axis as take_along_axis
from keras.src.ops.numpy import tan as tan
from keras.src.ops.numpy import tanh as tanh
from keras.src.ops.numpy import tensordot as tensordot
from keras.src.ops.numpy import tile as tile
from keras.src.ops.numpy import trace as trace
from keras.src.ops.numpy import transpose as transpose
from keras.src.ops.numpy import trapezoid as trapezoid
from keras.src.ops.numpy import tri as tri
from keras.src.ops.numpy import tril as tril
from keras.src.ops.numpy import triu as triu
from keras.src.ops.numpy import true_divide as true_divide
from keras.src.ops.numpy import trunc as trunc
from keras.src.ops.numpy import unravel_index as unravel_index
from keras.src.ops.numpy import vander as vander
from keras.src.ops.numpy import var as var
from keras.src.ops.numpy import vdot as vdot
from keras.src.ops.numpy import vectorize as vectorize
from keras.src.ops.numpy import view as view
from keras.src.ops.numpy import vsplit as vsplit
from keras.src.ops.numpy import vstack as vstack
from keras.src.ops.numpy import where as where
from keras.src.ops.numpy import zeros as zeros
from keras.src.ops.numpy import zeros_like as zeros_like



from keras.src.export.litert import LiteRTExporter
from keras.src.export.litert import export_litert
from keras.src.export.onnx import export_onnx
from keras.src.export.openvino import export_openvino
from keras.src.export.saved_model import ExportArchive
from keras.src.export.saved_model import export_saved_model
from keras.src.export.tfsm_layer import TFSMLayer



import numpy as np

from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.callbacks.callback import Callback
from keras.src.utils import io_utils


@keras_export("keras.callbacks.LearningRateScheduler")
class LearningRateScheduler(Callback):
    """Learning rate scheduler.

    At the beginning of every epoch, this callback gets the updated learning
    rate value from `schedule` function provided at `__init__`, with the current
    epoch and current learning rate, and applies the updated learning rate on
    the optimizer.

    Args:
        schedule: A function that takes an epoch index (integer, indexed from 0)
            and current learning rate (float) as inputs and returns a new
            learning rate as output (float).
        verbose: Integer. 0: quiet, 1: log update messages.

    Example:

    >>> # This function keeps the initial learning rate for the first ten epochs
    >>> # and decreases it exponentially after that.
    >>> def scheduler(epoch, lr):
    ...     if epoch < 10:
    ...         return lr
    ...     else:
    ...         return lr * ops.exp(-0.1)
    >>>
    >>> model = keras.models.Sequential([keras.layers.Dense(10)])
    >>> model.compile(keras.optimizers.SGD(), loss='mse')
    >>> round(model.optimizer.learning_rate, 5)
    0.01

    >>> callback = keras.callbacks.LearningRateScheduler(scheduler)
    >>> history = model.fit(np.arange(100).reshape(5, 20), np.zeros(5),
    ...                     epochs=15, callbacks=[callback], verbose=0)
    >>> round(model.optimizer.learning_rate, 5)
    0.00607

    """

    def __init__(self, schedule, verbose=0):
        super().__init__()
        self.schedule = schedule
        self.verbose = verbose

    def on_epoch_begin(self, epoch, logs=None):
        if not hasattr(self.model.optimizer, "learning_rate"):
            raise ValueError('Optimizer must have a "learning_rate" attribute.')

        try:  # new API
            learning_rate = float(
                backend.convert_to_numpy(self.model.optimizer.learning_rate)
            )
            learning_rate = self.schedule(epoch, learning_rate)
        except TypeError:  # Support for old API for backward compatibility
            learning_rate = self.schedule(epoch)

        if not isinstance(learning_rate, (float, np.float32, np.float64)):
            raise ValueError(
                "The output of the `schedule` function should be a float. "
                f"Got: {learning_rate}"
            )

        self.model.optimizer.learning_rate = learning_rate
        if self.verbose > 0:
            io_utils.print_msg(
                f"\nEpoch {epoch + 1}: LearningRateScheduler setting learning "
                f"rate to {learning_rate}."
            )

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs["learning_rate"] = float(
            backend.convert_to_numpy(self.model.optimizer.learning_rate)
        )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.activations import deserialize as deserialize
from keras.src.activations import get as get
from keras.src.activations import serialize as serialize
from keras.src.activations.activations import celu as celu
from keras.src.activations.activations import elu as elu
from keras.src.activations.activations import exponential as exponential
from keras.src.activations.activations import gelu as gelu
from keras.src.activations.activations import glu as glu
from keras.src.activations.activations import hard_shrink as hard_shrink
from keras.src.activations.activations import hard_sigmoid as hard_sigmoid
from keras.src.activations.activations import hard_silu as hard_silu
from keras.src.activations.activations import hard_silu as hard_swish
from keras.src.activations.activations import hard_tanh as hard_tanh
from keras.src.activations.activations import leaky_relu as leaky_relu
from keras.src.activations.activations import linear as linear
from keras.src.activations.activations import log_sigmoid as log_sigmoid
from keras.src.activations.activations import log_softmax as log_softmax
from keras.src.activations.activations import mish as mish
from keras.src.activations.activations import relu as relu
from keras.src.activations.activations import relu6 as relu6
from keras.src.activations.activations import selu as selu
from keras.src.activations.activations import sigmoid as sigmoid
from keras.src.activations.activations import silu as silu
from keras.src.activations.activations import silu as swish
from keras.src.activations.activations import soft_shrink as soft_shrink
from keras.src.activations.activations import softmax as softmax
from keras.src.activations.activations import softplus as softplus
from keras.src.activations.activations import softsign as softsign
from keras.src.activations.activations import sparse_plus as sparse_plus
from keras.src.activations.activations import sparse_sigmoid as sparse_sigmoid
from keras.src.activations.activations import sparsemax as sparsemax
from keras.src.activations.activations import squareplus as squareplus
from keras.src.activations.activations import tanh as tanh
from keras.src.activations.activations import tanh_shrink as tanh_shrink
from keras.src.activations.activations import threshold as threshold



from keras.src import backend


def is_in_jax_tracing_scope(x=None):
    if backend.backend() == "jax":
        import jax

        if x is None:
            x = backend.numpy.ones(())
        return isinstance(x, jax.core.Tracer)
    return False



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.dtype_policies import deserialize as deserialize
from keras.src.dtype_policies import get as get
from keras.src.dtype_policies import serialize as serialize
from keras.src.dtype_policies.dtype_policy import (
    AWQDTypePolicy as AWQDTypePolicy,
)
from keras.src.dtype_policies.dtype_policy import DTypePolicy as DTypePolicy
from keras.src.dtype_policies.dtype_policy import (
    FloatDTypePolicy as FloatDTypePolicy,
)
from keras.src.dtype_policies.dtype_policy import (
    GPTQDTypePolicy as GPTQDTypePolicy,
)
from keras.src.dtype_policies.dtype_policy import (
    Int4DTypePolicy as Int4DTypePolicy,
)
from keras.src.dtype_policies.dtype_policy import (
    QuantizedDTypePolicy as QuantizedDTypePolicy,
)
from keras.src.dtype_policies.dtype_policy import (
    QuantizedFloat8DTypePolicy as QuantizedFloat8DTypePolicy,
)
from keras.src.dtype_policies.dtype_policy_map import (
    DTypePolicyMap as DTypePolicyMap,
)



from keras.src.api_export import keras_export
from keras.src.quantizers.quantization_config import QuantizationConfig


@keras_export("keras.quantizers.GPTQConfig")
class GPTQConfig(QuantizationConfig):
    """Configuration class for the GPTQ (Gradient-based Post-Training
    Quantization) algorithm.

    GPTQ is a post-training quantization method that quantizes neural network
    weights to lower precision (e.g., 4-bit) while minimizing the impact on
    model accuracy. It works by analyzing the Hessian matrix of the loss
    function with respect to the weights and applying optimal quantization
    that preserves the most important weight values.

    **When to use GPTQ:**
    - You want to reduce model size and memory usage
    - You need faster inference on hardware that supports low-precision
      operations
    - You want to maintain model accuracy as much as possible
    - You have a pre-trained model that you want to quantize without
      retraining

    **How it works:**
    1. Uses calibration data to compute the Hessian matrix for each layer
    2. Applies iterative quantization with error correction
    3. Reorders weights based on activation importance (optional)
    4. Quantizes weights while minimizing quantization error

    **Example usage:**
    ```python
    from keras.quantizers import GPTQConfig
    from keras import Model

    # Create configuration for 4-bit quantization
    config = GPTQConfig(
        dataset=calibration_data,          # Your calibration dataset
        tokenizer=your_tokenizer,          # Tokenizer for text data
        weight_bits=4,                     # Quantize to 4 bits
        num_samples=128,                   # Number of calibration samples
        sequence_length=512,               # Sequence length for each sample
        hessian_damping=0.01,             # Hessian stabilization factor
        group_size=128,                    # Weight grouping for quantization
        symmetric=False,                   # Use asymmetric quantization
        activation_order=True              # Reorder weights by importance
    )

    # Apply quantization to your model
    model = Model(...)  # Your pre-trained model
    model.quantize("gptq", config=config)

    # The model now has quantized weights and can be used for inference
    ```

    **Benefits:**
    - **Memory reduction**: 4-bit quantization reduces memory by ~8x compared
      to float32
    - **Faster inference**: Lower precision operations are faster on supported
      hardware
    - **Accuracy preservation**: Minimizes accuracy loss through optimal
      quantization
    - **No retraining required**: Works with pre-trained models

    **Advanced usage examples:**

    **Per-channel quantization (recommended for most cases):**
    ```python
    config = GPTQConfig(
        dataset=calibration_data,
        tokenizer=tokenizer,
        weight_bits=4,
        group_size=-1,  # -1 enables per-channel quantization
        symmetric=False
    )
    ```

    **Grouped quantization (for specific hardware requirements):**
    ```python
    config = GPTQConfig(
        dataset=calibration_data,
        tokenizer=tokenizer,
        weight_bits=4,
        group_size=64,  # 64 weights share the same scale factor
        symmetric=True   # Use symmetric quantization
    )
    ```

    **High-accuracy quantization with activation ordering:**
    ```python
    config = GPTQConfig(
        dataset=calibration_data,
        tokenizer=tokenizer,
        weight_bits=4,
        activation_order=True,  # Reorder weights by importance
        hessian_damping=0.005,  # Lower damping for more precise
        # quantization
        num_samples=256          # More samples for better accuracy
    )
    ```

    **References:**
    - Original GPTQ paper: "GPTQ: Accurate Post-Training Quantization
      for Generative Pre-trained Transformers"
    - Implementation based on: https://github.com/IST-DASLab/gptq
    - Suitable for: Transformer models, large language models, and other
      deep neural networks

    **Note:** The quality of quantization depends heavily on the calibration
    dataset. Use representative data that covers the expected input
    distribution for best results.

    Args:
        dataset: The calibration dataset. It can be an iterable that yields
            strings or pre-tokenized numerical tensors (e.g., a list of
            strings, a generator, or a NumPy array). This data is used to
            analyze the model's activations.
        tokenizer: A `keras_nlp.Tokenizer` instance (or a similar callable)
            that is used to process the `dataset` if it contains strings.
        weight_bits: (int, optional) The number of bits to quantize weights to.
            Defaults to 4.
        num_samples: (int, optional) The number of calibration data samples to
            use from the dataset. Defaults to 128.
        sequence_length: (int, optional) The sequence length to use for each
            calibration sample. Defaults to 512.
        hessian_damping: (float, optional) The % of Hessian damping to use for
            stabilization during inverse calculation. Defaults to 0.01.
        group_size: (int, optional) The size of weight groups to quantize
            together. A `group_size` of -1 indicates per-channel quantization.
            Defaults to 128.
        symmetric: (bool, optional) If `True`, uses symmetric quantization.
            If `False`, uses asymmetric quantization. Defaults to `False`.
        activation_order: (bool, optional) If `True`, reorders weight columns
            based on activation magnitude, which can improve quantization
            accuracy. Defaults to `False`.
        quantization_layer_structure: (dict, optional) A dictionary defining the
            model's quantization structure. It should contain:
            - "pre_block_layers": list of layers to run before the first block.
            - "sequential_blocks": list of blocks to be quantized sequentially.
            If not provided, the model must implement
            `get_quantization_layer_structure`.
    """

    def __init__(
        self,
        dataset,
        tokenizer,
        *,
        weight_bits: int = 4,
        num_samples: int = 128,
        per_channel: bool = True,
        sequence_length: int = 512,
        hessian_damping: float = 0.01,
        group_size: int = 128,
        symmetric: bool = False,
        activation_order: bool = False,
        quantization_layer_structure: dict = None,
    ):
        super().__init__()
        if weight_bits not in [2, 3, 4, 8]:
            raise ValueError(
                f"Unsupported weight_bits {weight_bits}. "
                "Supported values are 2, 3, 4, and 8."
            )
        if num_samples <= 0:
            raise ValueError("num_samples must be a positive integer.")
        if sequence_length <= 0:
            raise ValueError("sequence_length must be a positive integer.")
        if hessian_damping < 0 or hessian_damping > 1:
            raise ValueError("hessian_damping must be between 0 and 1.")
        if group_size < -1 or group_size == 0:
            raise ValueError(
                "Invalid group_size. Supported values are -1 (whole-tensor) "
                "or a positive integer, "
                f"but got {group_size}."
            )
        self.dataset = dataset
        self.tokenizer = tokenizer
        self.num_samples = num_samples
        self.per_channel = per_channel
        self.sequence_length = sequence_length
        self.hessian_damping = hessian_damping
        self.weight_bits = weight_bits
        self.group_size = group_size
        self.symmetric = symmetric
        self.activation_order = activation_order
        self.quantization_layer_structure = quantization_layer_structure

    def get_config(self):
        return {
            # Dataset and Tokenizer are only required for a one-time
            # calibration and are not saved in the config.
            "dataset": None,
            "tokenizer": None,
            "weight_bits": self.weight_bits,
            "num_samples": self.num_samples,
            "per_channel": self.per_channel,
            "sequence_length": self.sequence_length,
            "hessian_damping": self.hessian_damping,
            "group_size": self.group_size,
            "symmetric": self.symmetric,
            "activation_order": self.activation_order,
            "quantization_layer_structure": self.quantization_layer_structure,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    @property
    def mode(self):
        return "gptq"

    def dtype_policy_string(self):
        """Returns the dtype policy string for this configuration.

        Returns:
            A string representing the dtype policy, e.g. "gptq_4bit".
        """
        return f"gptq/{self.weight_bits}/{self.group_size}"



import collections
import collections.abc
import itertools

from keras.src.backend.config import backend
from keras.src.utils.module_utils import dmtree

# NOTE: There are two known discrepancies between this `dmtree` implementation
# of the tree API and the `optree` implementation:
#
# 1. `map_structure` with *multiple* structures and `map_structure_up_to` do not
#    use the object registration (they use the raw `dmtree.map_structure` and
#    `dmtree.map_structure_up_to`). This only has consequences with two types of
#    structures:
#    - `TrackedSet` will not explored (considered as a leaf).
#    - `OrderedDict` will be traversed in the order of sorted keys, not the
#      order of the items. This is typically inconsequential because functions
#      used with `map_structure` and `map_structure_up_to` are typically not
#      order dependent and are, in fact, stateless.
#
# 2. The handling of non-sortable keys in dictionaries in inconsistent. `optree`
#    uses the iteration order while `dmtree` raises an error. This is not an
#    issue as keys are always strings. But this is the reason why we document
#    non-sortable keys as unsupported (meaning behavior is undefined).

REGISTERED_CLASSES = {}

ClassRegistration = collections.namedtuple(
    "ClassRegistration", ["flatten", "unflatten"]
)


class TypeErrorRemapping:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is TypeError:
            raise ValueError(exc_value).with_traceback(traceback)
        return False


def register_tree_node(
    cls,
    flatten_func=None,
    unflatten_func=None,
):
    if flatten_func is None:
        flatten_func = lambda x: x.tree_flatten()
    if unflatten_func is None:
        unflatten_func = cls.tree_unflatten
    REGISTERED_CLASSES[cls] = ClassRegistration(flatten_func, unflatten_func)


def register_tree_node_class(cls):
    register_tree_node(cls)
    return cls


register_tree_node(
    collections.OrderedDict,
    lambda d: (d.values(), list(d.keys()), d.keys()),
    lambda metadata, children: collections.OrderedDict(zip(metadata, children)),
)

if backend() == "tensorflow":
    from tensorflow.python.trackable.data_structures import ListWrapper
    from tensorflow.python.trackable.data_structures import _DictWrapper

    register_tree_node(
        ListWrapper,
        lambda x: (x, None),
        lambda metadata, children: ListWrapper(list(children)),
    )

    def sorted_keys_and_values(d):
        keys = sorted(list(d.keys()))
        values = [d[k] for k in keys]
        return values, keys, keys

    register_tree_node(
        _DictWrapper,
        sorted_keys_and_values,
        lambda metadata, children: _DictWrapper(
            {key: child for key, child in zip(metadata, children)}
        ),
    )


def is_nested(structure):
    return type(structure) in REGISTERED_CLASSES or dmtree.is_nested(structure)


def traverse(func, structure, top_down=True):
    if not callable(func):
        raise TypeError(
            f"`func` must be callable, got {func} of type {type(func)}"
        )

    def remap_map_to_none(value, new_value):
        if isinstance(value, type) and value.__name__ == "MAP_TO_NONE":
            return new_value
        return value

    def traverse_top_down(s):
        ret = func(s)
        if ret is not None:
            return remap_map_to_none(ret, dmtree.MAP_TO_NONE)
        registration = REGISTERED_CLASSES.get(type(s), None)
        if registration is None:
            return None
        flat_meta_s = registration.flatten(s)
        flat_s = [
            dmtree.traverse(traverse_top_down, x, top_down=True)
            for x in list(flat_meta_s[0])
        ]
        return registration.unflatten(flat_meta_s[1], flat_s)

    def traverse_bottom_up(s):
        registration = REGISTERED_CLASSES.get(type(s), None)
        if registration is not None:
            flat_meta_s = registration.flatten(s)
            ret = [traverse_bottom_up(x) for x in list(flat_meta_s[0])]
            ret = registration.unflatten(flat_meta_s[1], ret)
        elif not dmtree.is_nested(s):
            ret = s
        elif isinstance(s, collections.abc.Mapping):
            ret = [traverse_bottom_up(s[key]) for key in sorted(s)]
            ret = dmtree._sequence_like(s, ret)
        else:
            ret = [traverse_bottom_up(x) for x in s]
            ret = dmtree._sequence_like(s, ret)
        func_ret = func(ret)
        return ret if func_ret is None else remap_map_to_none(func_ret, None)

    if top_down:
        return dmtree.traverse(traverse_top_down, structure, top_down=True)
    else:
        return traverse_bottom_up(structure)


def flatten(structure):
    if not is_nested(structure):
        return [structure]

    flattened = []

    def flatten_func(s):
        registration = REGISTERED_CLASSES.get(type(s), None)
        if registration is not None:
            flat_s = list(registration.flatten(s)[0])
            return dmtree.traverse(flatten_func, flat_s, top_down=True)
        if not is_nested(s):
            flattened.append(s)
            return dmtree.MAP_TO_NONE if s is None else s
        return None

    dmtree.traverse(flatten_func, structure, top_down=True)
    return flattened


def _recursive_flatten_with_path(path, structure, flattened):
    registration = REGISTERED_CLASSES.get(type(structure), None)
    if registration is not None:
        flat_meta_paths = registration.flatten(structure)
        flat = flat_meta_paths[0]
        paths = (
            flat_meta_paths[2]
            if len(flat_meta_paths) >= 3
            else itertools.count()
        )
        for key, value in zip(paths, flat):
            _recursive_flatten_with_path(path + (key,), value, flattened)
    elif not dmtree.is_nested(structure):
        flattened.append((path, structure))
    elif isinstance(structure, collections.abc.Mapping):
        for key in sorted(structure):
            _recursive_flatten_with_path(
                path + (key,), structure[key], flattened
            )
    else:
        for key, value in enumerate(structure):
            _recursive_flatten_with_path(path + (key,), value, flattened)


def flatten_with_path(structure):
    if not is_nested(structure):
        return [((), structure)]

    # Fully reimplemented in Python to handle registered classes, OrderedDict
    # and namedtuples the same way as optree.
    flattened = []
    _recursive_flatten_with_path((), structure, flattened)
    return flattened


def map_structure(func, *structures, none_is_leaf=True):
    if not callable(func):
        raise TypeError(
            f"`func` must be callable, got {func} of type {type(func)}"
        )

    map_func = func
    if not none_is_leaf:

        def func_skipping_none(*args):
            # Check if the reference entry (first one) is None
            if args[0] is None:
                if not all(s is None for s in args):
                    raise ValueError(
                        "Structure mismatch: some arguments are None, others "
                        f"are not. Received arguments: {args}."
                    )
                return None
            return func(*args)

        map_func = func_skipping_none

    def func_traverse_wrapper(s):
        if is_nested(s):
            return None
        ret = map_func(s)
        if ret is None:
            return dmtree.MAP_TO_NONE
        return ret

    if len(structures) == 1:
        return traverse(func_traverse_wrapper, structures[0])

    with TypeErrorRemapping():
        return dmtree.map_structure(map_func, *structures)


def map_structure_up_to(shallow_structure, func, *structures):
    if not callable(func):
        raise TypeError(
            f"`func` must be callable, got {func} of type {type(func)}"
        )

    with TypeErrorRemapping():
        return dmtree.map_structure_up_to(shallow_structure, func, *structures)


def assert_same_structure(a, b):
    # Fully reimplemented in Python to handle registered classes.

    # Don't handle OrderedDict as a registered class, use the normal dict path
    # so that OrderedDict is equivalent to dict per optree behavior.
    a_registration = REGISTERED_CLASSES.get(type(a), None)
    if type(a) is collections.OrderedDict:
        a_registration = None

    b_registration = REGISTERED_CLASSES.get(type(b), None)
    if type(b) is collections.OrderedDict:
        b_registration = None

    if a_registration != b_registration:
        raise ValueError(
            f"Custom node type mismatch; "
            f"expected type: {type(a)}, got type: {type(b)} "
            f"while comparing {a} and {b}."
        )
    if a_registration is not None:
        a_flat_meta = a_registration.flatten(a)
        b_flat_meta = b_registration.flatten(b)
        a_flat = list(a_flat_meta[0])
        b_flat = list(b_flat_meta[0])
        if not a_flat_meta[1] == b_flat_meta[1]:
            raise ValueError(
                f"Mismatch custom node data; "
                f"expected: {a_flat_meta[1]}, got: {b_flat_meta[1]} "
                f"while comparing {a} and {b}."
            )
        if len(a_flat) != len(b_flat):
            raise ValueError(
                f"Arity mismatch; expected: {len(a)}, got: {len(b)} "
                f"while comparing {a} and {b}."
            )
        for sub_a, sub_b in zip(a_flat, b_flat):
            assert_same_structure(sub_a, sub_b)
    elif not dmtree.is_nested(a):
        if dmtree.is_nested(b):
            raise ValueError(
                f"Structures don't have the same nested structure: {a}, {b}."
            )
    elif isinstance(
        a, (dict, collections.OrderedDict, collections.defaultdict)
    ):
        if not isinstance(
            b, (dict, collections.OrderedDict, collections.defaultdict)
        ):
            raise ValueError(
                f"Expected an instance of dict, collections.OrderedDict, or "
                f"collections.defaultdict, got {type(b)} "
                f"while comparing {a} and {b}."
            )
        a_keys = sorted(a)
        b_keys = sorted(b)
        if not a_keys == b_keys:
            raise ValueError(
                f"Dictionary key mismatch; "
                f"expected key(s): {a_keys}, got key(s): {b_keys} "
                f"while comparing {a} and {b}."
            )
        for key in a_keys:
            assert_same_structure(a[key], b[key])
    elif isinstance(a, collections.abc.Mapping):
        raise ValueError(
            f"Encountered unregistered collections.abc.Mapping type: {type(a)} "
            f"while comparing {a} and {b}."
        )
    else:
        if type(a) is not type(b):
            raise ValueError(
                f"Expected an instance of {type(a)}, got {type(b)} "
                f"while comparing {a} and {b}."
            )
        if not len(a) == len(b):
            raise ValueError(
                f"Arity mismatch; expected: {len(a)}, got: {len(b)} "
                f"while comparing {a} and {b}."
            )
        for sub_a, sub_b in zip(a, b):
            assert_same_structure(sub_a, sub_b)


def assert_same_paths(a, b):
    a_paths = set([path for path, _ in flatten_with_path(a)])
    b_paths = set([path for path, _ in flatten_with_path(b)])

    if a_paths != b_paths:
        msg = "`a` and `b` don't have the same paths."
        a_diff = a_paths.difference(b_paths)
        if a_diff:
            msg += f"\nPaths in `a` missing in `b`:\n{a_diff}"
        b_diff = b_paths.difference(a_paths)
        if b_diff:
            msg += f"\nPaths in `b` missing in `a`:\n{b_diff}"
        raise ValueError(msg)


def pack_sequence_as(structure, flat_sequence):
    # This is not just an optimization for the case when structure is a leaf.
    # This is required to avoid Torch Dynamo failures.
    if not is_nested(structure):
        if len(flat_sequence) == 1:
            return flat_sequence[0]
        else:
            raise ValueError(
                "Incorrect number of leaves provided by `flat_sequence` for "
                f"`structure`; expected: 1, got {len(flat_sequence)}."
            )

    flat_sequence_it = enumerate(flat_sequence)

    def unflatten_func(s):
        registration = REGISTERED_CLASSES.get(type(s), None)
        if registration is not None:
            flat_meta_s = registration.flatten(s)
            flat_s = dmtree.traverse(
                unflatten_func, list(flat_meta_s[0]), top_down=True
            )
            return registration.unflatten(flat_meta_s[1], flat_s)
        elif not dmtree.is_nested(s):
            try:
                _, value = next(flat_sequence_it)
                return dmtree.MAP_TO_NONE if value is None else value
            except StopIteration:
                raise ValueError(
                    "Too few leaves provided by `flat_sequence` for "
                    f"`structure`. Got {len(flat_sequence)}."
                )
        return None

    ret = dmtree.traverse(unflatten_func, structure, top_down=True)
    try:
        index, _ = next(flat_sequence_it)
        raise ValueError(
            "Too many leaves provided by `flat_sequence` for `structure`; "
            f"expected: {index}, got {len(flat_sequence)}."
        )
    except StopIteration:
        return ret


def lists_to_tuples(structure):
    def list_to_tuple(instance):
        return tuple(instance) if isinstance(instance, list) else None

    return traverse(list_to_tuple, structure, top_down=False)


def map_shape_structure(func, structure):
    if not callable(func):
        raise TypeError(
            f"`func` must be callable, got {func} of type {type(func)}"
        )

    def map_shape_func(x):
        if isinstance(x, (list, tuple)) and all(
            isinstance(e, (int, type(None))) for e in x
        ):
            ret = func(x)
        elif is_nested(x):
            return None
        else:
            ret = func(x)
        return ret if ret is not None else dmtree.MAP_TO_NONE

    return traverse(map_shape_func, structure, top_down=True)



import optree
import optree.utils

from keras.src.backend.config import backend


def register_tree_node_class(cls):
    return optree.register_pytree_node_class(cls, namespace="keras")


# Register backend-specific node classes
if backend() == "tensorflow":
    from tensorflow.python.trackable.data_structures import ListWrapper
    from tensorflow.python.trackable.data_structures import _DictWrapper

    try:
        optree.register_pytree_node(
            ListWrapper,
            lambda x: (x, None),
            lambda metadata, children: ListWrapper(list(children)),
            namespace="keras",
        )

        def sorted_keys_and_values(d):
            keys = sorted(list(d.keys()))
            values = [d[k] for k in keys]
            return values, keys, keys

        optree.register_pytree_node(
            _DictWrapper,
            sorted_keys_and_values,
            lambda metadata, children: _DictWrapper(
                {key: child for key, child in zip(metadata, children)}
            ),
            namespace="keras",
        )
    except ValueError:
        pass  # We may have already registered if we are reimporting keras.


def is_nested(structure):
    return not optree.tree_is_leaf(
        structure, none_is_leaf=True, namespace="keras"
    )


def traverse(func, structure, top_down=True):
    # From https://github.com/google/jax/pull/19695
    def traverse_children():
        children, treedef = optree.tree_flatten(
            structure,
            is_leaf=lambda x: x is not structure,
            none_is_leaf=True,
            namespace="keras",
        )
        if treedef.num_nodes == 1 and treedef.num_leaves == 1:
            return structure
        else:
            return optree.tree_unflatten(
                treedef,
                [traverse(func, c, top_down=top_down) for c in children],
            )

    if top_down:
        ret = func(structure)
        if ret is None:
            return traverse_children()
    else:
        traversed_structure = traverse_children()
        ret = func(traversed_structure)
        if ret is None:
            return traversed_structure
    # Detect MAP_TO_NONE without tree_api import to avoid circular import.
    if isinstance(ret, type) and ret.__name__ == "MAP_TO_NONE":
        return None
    return ret


def flatten(structure):
    # optree.tree_flatten returns a pair (leaves, treespec) where the first
    # element is a list of leaf values and the second element is a treespec
    # representing the structure of the pytree.
    leaves, _ = optree.tree_flatten(
        structure, none_is_leaf=True, namespace="keras"
    )
    return leaves


def flatten_with_path(structure):
    paths, leaves, _ = optree.tree_flatten_with_path(
        structure, none_is_leaf=True, namespace="keras"
    )
    return list(zip(paths, leaves))


def map_structure(func, *structures, none_is_leaf=True):
    if not structures:
        raise ValueError("Must provide at least one structure")

    # Add check for same structures, otherwise optree just maps to shallowest.
    def func_with_check(*args):
        if not all(
            optree.tree_is_leaf(s, none_is_leaf=none_is_leaf, namespace="keras")
            for s in args
        ):
            raise ValueError("Structures don't have the same nested structure.")
        return func(*args)

    map_func = func_with_check if len(structures) > 1 else func

    return optree.tree_map(
        map_func, *structures, none_is_leaf=none_is_leaf, namespace="keras"
    )


def map_structure_up_to(shallow_structure, func, *structures):
    if not structures:
        raise ValueError("Must provide at least one structure")

    # Add check that `shallow_structure` really is the shallowest.
    # Also only call `func` on `structures` and not `shallow_structure`.
    def func_with_check_without_shallow_structure(shallow, *args):
        if not optree.tree_is_leaf(shallow):
            raise ValueError("Structures don't have the same nested structure.")
        return func(*args)

    return optree.tree_map(
        func_with_check_without_shallow_structure,
        shallow_structure,
        *structures,
        none_is_leaf=True,
        namespace="keras",
    )


def assert_same_structure(a, b):
    def check(a_leaf, b_leaf):
        if not optree.tree_is_leaf(
            a_leaf, none_is_leaf=True, namespace="keras"
        ) or not optree.tree_is_leaf(
            b_leaf, none_is_leaf=True, namespace="keras"
        ):
            raise ValueError("Structures don't have the same nested structure.")
        return None

    optree.tree_map(check, a, b, none_is_leaf=True, namespace="keras")


def assert_same_paths(a, b):
    a_paths = set(optree.tree_paths(a, none_is_leaf=True, namespace="keras"))
    b_paths = set(optree.tree_paths(b, none_is_leaf=True, namespace="keras"))

    if a_paths != b_paths:
        msg = "`a` and `b` don't have the same paths."
        a_diff = a_paths.difference(b_paths)
        if a_diff:
            msg += f"\nPaths in `a` missing in `b`:\n{a_diff}"
        b_diff = b_paths.difference(a_paths)
        if b_diff:
            msg += f"\nPaths in `b` missing in `a`:\n{b_diff}"
        raise ValueError(msg)


def pack_sequence_as(structure, flat_sequence):
    _, treespec = optree.tree_flatten(
        structure, none_is_leaf=True, namespace="keras"
    )
    return optree.tree_unflatten(treespec, flat_sequence)


def lists_to_tuples(structure):
    def list_to_tuple(instance):
        return tuple(instance) if isinstance(instance, list) else None

    return traverse(list_to_tuple, structure, top_down=False)


def map_shape_structure(func, structure):
    def is_shape_tuple(x):
        return isinstance(x, (list, tuple)) and all(
            isinstance(e, (int, type(None))) for e in x
        )

    return optree.tree_map(
        func,
        structure,
        is_leaf=is_shape_tuple,
        none_is_leaf=True,
        namespace="keras",
    )



from keras.src import activations
from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.layers.layer import Layer


def _large_negative_number(dtype):
    """Return a Large negative number based on dtype."""
    if backend.standardize_dtype(dtype) == "float16":
        return -3e4
    return -1e9


@keras_export("keras.layers.Softmax")
class Softmax(Layer):
    """Softmax activation layer.

    Formula:
    ``` python
    exp_x = exp(x - max(x))
    f(x) = exp_x / sum(exp_x)
    ```

    Example:
    >>> softmax_layer = keras.layers.Softmax()
    >>> input = np.array([1.0, 2.0, 1.0])
    >>> result = softmax_layer(input)
    >>> result
    [0.21194157, 0.5761169, 0.21194157]


    Args:
        axis: Integer, or list of Integers, axis along which the softmax
            normalization is applied.
        **kwargs: Base layer keyword arguments, such as `name` and `dtype`.

    Call arguments:
        inputs: The inputs (logits) to the softmax layer.
        mask: A boolean mask that is broadcastable to `inputs`. The mask
            specifies 1 to keep and 0 to mask. Each dimension of the mask
            must either be 1 or match the corresponding dimension of
            `inputs`; it must not be larger. Defaults to `None`.

    Returns:
        Softmaxed output with the same shape as `inputs`.
    """

    def __init__(self, axis=-1, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis
        self.supports_masking = True

        self._build_at_init()

    def call(self, inputs, mask=None):
        if mask is not None:
            if len(mask.shape) > len(inputs.shape):
                raise ValueError(
                    "The `mask` must be broadcastable to `inputs` "
                    "and must not have more dimensions. "
                    f"Received: inputs.shape={inputs.shape}, "
                    f"mask.shape={mask.shape}"
                )
            for m_dim, i_dim in zip(mask.shape[::-1], inputs.shape[::-1]):
                if m_dim is not None and i_dim is not None:
                    if m_dim != 1 and m_dim != i_dim:
                        raise ValueError(
                            "The `mask` must be broadcastable to "
                            "`inputs`. Each mask dimension must be 1 "
                            "or match the corresponding input "
                            "dimension. Received: "
                            f"inputs.shape={inputs.shape}, "
                            f"mask.shape={mask.shape}"
                        )
            # We keep the positions where the mask is True or > 0.5, and set the
            # other (masked) positions to -1e.9.
            if backend.standardize_dtype(mask.dtype) != "bool":
                mask = backend.numpy.greater(
                    mask, backend.cast(0.5, dtype=mask.dtype)
                )
            inputs = backend.numpy.where(
                mask, inputs, _large_negative_number(inputs.dtype)
            )
        if isinstance(self.axis, (tuple, list)):
            if len(self.axis) > 1:
                outputs = backend.numpy.exp(
                    inputs
                    - backend.math.logsumexp(
                        inputs, axis=self.axis, keepdims=True
                    )
                )
            else:
                outputs = activations.softmax(inputs, axis=self.axis[0])
        else:
            outputs = activations.softmax(inputs, axis=self.axis)

        # Free pre-softmax masked inputs to reduce peak memory.
        # Without this, the masked inputs, softmax outputs, and
        # post-masked outputs all exist simultaneously.
        del inputs

        if mask is not None:
            # Zero out masked positions in case the entire axis is masked
            # (where softmax would output a uniform distribution).
            outputs = backend.numpy.where(mask, outputs, 0.0)

        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({"axis": self.axis})
        return config

    def compute_output_shape(self, input_shape):
        return input_shape



from keras.src.backend.config import backend

if backend() == "torch":
    # When using the torch backend,
    # torch needs to be imported first, otherwise it will segfault
    # upon import.
    import torch

from keras.src.api_export import keras_export
from keras.src.backend.common.dtypes import result_type
from keras.src.backend.common.keras_tensor import KerasTensor
from keras.src.backend.common.keras_tensor import any_symbolic_tensors
from keras.src.backend.common.keras_tensor import is_keras_tensor
from keras.src.backend.common.masking import get_keras_mask
from keras.src.backend.common.masking import set_keras_mask
from keras.src.backend.common.stateless_scope import StatelessScope
from keras.src.backend.common.stateless_scope import get_stateless_scope
from keras.src.backend.common.stateless_scope import in_stateless_scope
from keras.src.backend.common.symbolic_scope import SymbolicScope
from keras.src.backend.common.symbolic_scope import in_symbolic_scope
from keras.src.backend.common.variables import AutocastScope
from keras.src.backend.common.variables import Variable
from keras.src.backend.common.variables import get_autocast_scope
from keras.src.backend.common.variables import is_float_dtype
from keras.src.backend.common.variables import is_int_dtype
from keras.src.backend.common.variables import standardize_dtype
from keras.src.backend.common.variables import standardize_shape
from keras.src.backend.config import epsilon
from keras.src.backend.config import floatx
from keras.src.backend.config import image_data_format
from keras.src.backend.config import set_epsilon
from keras.src.backend.config import set_floatx
from keras.src.backend.config import set_image_data_format
from keras.src.backend.config import standardize_data_format

# Import backend functions.
if backend() == "tensorflow":
    from keras.src.backend.tensorflow import *  # noqa: F403
    from keras.src.backend.tensorflow.core import Variable as BackendVariable
elif backend() == "jax":
    from keras.src.backend.jax import *  # noqa: F403
    from keras.src.backend.jax.core import Variable as BackendVariable
elif backend() == "torch":
    from keras.src.backend.torch import *  # noqa: F403
    from keras.src.backend.torch.core import Variable as BackendVariable

    distribution_lib = None
elif backend() == "numpy":
    from keras.src.backend.numpy import *  # noqa: F403
    from keras.src.backend.numpy.core import Variable as BackendVariable

    distribution_lib = None
elif backend() == "openvino":
    from keras.src.backend.openvino import *  # noqa: F403
    from keras.src.backend.openvino.core import Variable as BackendVariable

    distribution_lib = None
else:
    raise ValueError(f"Unable to import backend : {backend()}")


@keras_export("keras.Variable")
class Variable(BackendVariable):  # noqa: F811
    pass


backend_name_scope = name_scope  # noqa: F405


@keras_export("keras.name_scope")
class name_scope(backend_name_scope):
    pass


@keras_export("keras.device")
def device(device_name):
    """Context manager for backend-agnostic device placement.

    Use this context manager to control on which device operations are performed
    and tensors are allocated. This works across all backends (TensorFlow, JAX,
    PyTorch). This is useful for memory management, data preprocessing, and
    multi-device setups.

    Args:
        device_name: String specifying the device in format
            `"device_type:device_index"`. For example: `"cpu:0"`, `"gpu:0"`,
            `"gpu:1"`. For the PyTorch backend, `"gpu"` is automatically
            converted to `"cuda"`.

    Example:

    Basic usage with CPU and GPU:

    ```python
    # Allocate tensors on CPU
    with keras.device("cpu:0"):
        cpu_tensor = keras.ops.ones((2, 2))

    # Allocate tensors on GPU (if available)
    with keras.device("gpu:0"):
        gpu_tensor = keras.ops.ones((2, 2))
    ```

    Practical example with CPU preprocessing and GPU training:

    ```python
    # Create dummy data and model
    x_raw = np.random.rand(128, 784)
    y_train = np.random.randint(0, 10, size=(128,))
    model = keras.Sequential([
        keras.Input(shape=(784,)),
        keras.layers.Dense(10)
    ])
    model.compile(
        optimizer="adam",
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True)
    )

    # Preprocess data on CPU
    with keras.device("cpu:0"):
        x_processed = keras.ops.cast(x_raw, "float32")

    # Train on GPU (if available)
    with keras.device("gpu:0"):
        model.fit(x_processed, y_train, epochs=2)
    ```

    Use cases:

    - **Memory management**: Keep large tensors on CPU to save GPU memory
    - **Data preprocessing**: Process data on CPU before training on GPU
    - **GPU / TPU setups**: Control what runs on GPU / TPU vs CPU
    - **Multi-device setups**: Control which device receives which tensors

    Device naming conventions:

    - `"cpu:0"` - First CPU
    - `"gpu:0"` - First GPU (works across all backends)
    - `"gpu:1"` - Second GPU

    Note: For distributed training across multiple devices, see the
    [distributed training guides](https://keras.io/guides/distributed_training/).
    """
    return device_scope(device_name)  # noqa: F405



import copy
import json

try:
    import difflib
except ImportError:
    difflib = None

from keras.src.api_export import keras_export


@keras_export("keras.utils.Config")
class Config:
    """A Config is a dict-like container for named values.

    It offers a few advantages over a plain dict:

    - Setting and retrieving values via attribute setting / getting.
    - Ability to freeze the config to ensure no accidental config modifications
        occur past a certain point in your program.
    - Easy serialization of the whole config as JSON.

    Examples:

    ```python
    # Create a config via constructor arguments
    config = Config("learning_rate"=0.1, "momentum"=0.9)

    # Then keep adding to it via attribute-style setting
    config.use_ema = True
    config.ema_overwrite_frequency = 100

    # You can also add attributes via dict-like access
    config["seed"] = 123

    # You can retrieve entries both via attribute-style
    # access and dict-style access
    assert config.seed == 100
    assert config["learning_rate"] == 0.1
    ```

    A config behaves like a dict:

    ```python
    config = Config("learning_rate"=0.1, "momentum"=0.9)
    for k, v in config.items():
        print(f"{k}={v}")

    print(f"keys: {list(config.keys())}")
    print(f"values: {list(config.values())}")
    ```

    In fact, it can be turned into one:

    ```python
    config = Config("learning_rate"=0.1, "momentum"=0.9)
    dict_config = config.as_dict()
    ```

    You can easily serialize a config to JSON:

    ```python
    config = Config("learning_rate"=0.1, "momentum"=0.9)

    json_str = config.to_json()
    ```

    You can also freeze a config to prevent further changes:

    ```python
    config = Config()
    config.optimizer = "adam"
    config.seed = 123

    # Freeze the config to prevent changes.
    config.freeze()
    assert config.frozen

    config.foo = "bar"  # This will raise an error.
    ```
    """

    __attrs__ = None

    def __init__(self, **kwargs):
        self._config = kwargs
        self._frozen = False
        self.__attrs__ = set(dir(self))

    @property
    def frozen(self):
        """Returns True if the config is frozen."""
        return self._frozen

    def freeze(self):
        """Marks the config as frozen, preventing any ulterior modification."""
        self._frozen = True

    def unfreeze(self):
        self._frozen = False

    def _raise_if_frozen(self):
        if self._frozen:
            raise ValueError(
                "Cannot mutate attribute(s) because the config is frozen."
            )

    def as_dict(self):
        return copy.copy(self._config)

    def to_json(self):
        return json.dumps(self._config)

    def keys(self):
        return self._config.keys()

    def values(self):
        return self._config.values()

    def items(self):
        return self._config.items()

    def pop(self, *args):
        self._raise_if_frozen()
        return self._config.pop(*args)

    def update(self, *args, **kwargs):
        self._raise_if_frozen()
        return self._config.update(*args, **kwargs)

    def get(self, keyname, value=None):
        return self._config.get(keyname, value)

    def __setattr__(self, name, value):
        attrs = object.__getattribute__(self, "__attrs__")
        if attrs is None or name in attrs:
            return object.__setattr__(self, name, value)

        self._raise_if_frozen()
        self._config[name] = value

    def __getattr__(self, name):
        attrs = object.__getattribute__(self, "__attrs__")
        if attrs is None or name in attrs:
            return object.__getattribute__(self, name)

        if name in self._config:
            return self._config[name]

        msg = f"Unknown attribute: '{name}'."
        if difflib is not None:
            closest_matches = difflib.get_close_matches(
                name, self._config.keys(), n=1, cutoff=0.7
            )
            if closest_matches:
                msg += f" Did you mean '{closest_matches[0]}'?"
        raise AttributeError(msg)

    def __setitem__(self, key, item):
        self._raise_if_frozen()
        self._config[key] = item

    def __getitem__(self, key):
        return self._config[key]

    def __repr__(self):
        return f"<Config {self._config}>"

    def __iter__(self):
        keys = sorted(self._config.keys())
        for k in keys:
            yield k

    def __len__(self):
        return len(self._config)

    def __delitem__(self, key):
        self._raise_if_frozen()
        del self._config[key]

    def __contains__(self, item):
        return item in self._config
